import os, textwrap
import numpy as np
import unittest
from context import dpti
# from potential_common import soft_param, soft_param_three_element, meam_model
# from dpti.lib.lammps import get_natoms, get_thermo, get_last_dump
# from dpti.lib.dump import from_system_data
from potential_common import soft_param, meam_model
from dpti.hti_liq import _gen_lammps_input_ideal
from numpy.testing import assert_almost_equal
from unittest.mock import MagicMock, patch, PropertyMock

class TestGenLammpsIdeal(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    @patch('numpy.random')
    def test_soft_on(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(step='soft_on',
                    conf_file='conf.lmp', 
                    mass_map=[118.71,],
                    lamb=0.075,
                    soft_param=soft_param,
                    model='graph.pb',
                    nsteps=500000,
                    timestep=0.002,
                    ens='npt',
                    temp=1200,
                    pres=1.0, 
                    tau_t=0.1,
                    tau_p=0.5,
                    thermo_freq=100, 
                    copies=None,
                    norm_style='first',
                    if_meam=False,
                    meam_model=None
                )
        
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        TEMP            equal 1200.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        ZERO            equal 0
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        pair_style      lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      1 1 ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        compute         allmsd all msd
        thermo          ${THERMO_FREQ}
        thermo_style    custom step ke pe etotal enthalpy temp press vol c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input_ideal(**input)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_deep_on(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(step='deep_on',
                    conf_file='conf.lmp', 
                    mass_map=[118.71,],
                    lamb=0.075,
                    soft_param=soft_param,
                    model='graph.pb',
                    nsteps=500000,
                    timestep=0.002,
                    ens='npt',
                    temp=1200,
                    pres=1.0, 
                    tau_t=0.1,
                    tau_p=0.5,
                    thermo_freq=100, 
                    copies=None,
                    norm_style='first',
                    if_meam=False,
                    meam_model=None
                )
        
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        TEMP            equal 1200.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        ZERO            equal 0
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        compute         allmsd all msd
        thermo          ${THERMO_FREQ}
        thermo_style    custom step ke pe etotal enthalpy temp press vol c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input_ideal(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)


    @patch('numpy.random')
    def test_soft_off(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(step='soft_off',
                    conf_file='conf.lmp', 
                    mass_map=[118.71,],
                    lamb=0.075,
                    soft_param=soft_param,
                    model='graph.pb',
                    nsteps=500000,
                    timestep=0.002,
                    ens='npt',
                    temp=1200,
                    pres=1.0, 
                    tau_t=0.1,
                    tau_p=0.5,
                    thermo_freq=100, 
                    copies=None,
                    norm_style='first',
                    if_meam=False,
                    meam_model=None
                )
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        TEMP            equal 1200.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        ZERO            equal 0
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        INV_LAMBDA equal 1-${LAMBDA}
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        compute         allmsd all msd
        thermo          ${THERMO_FREQ}
        thermo_style    custom step ke pe etotal enthalpy temp press vol c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input_ideal(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)




    # @patch('numpy.random')
    # def test_deepmd_deep_on(self, patch_random):
    #     patch_random.randint = MagicMock(return_value=7858)
    #     input = dict ( conf_file='conf.lmp', 
    #                     mass_map=[118.71],
    #                     lamb=0.075,
    #                     model="graph.pb",
    #                     m_spring_k=[2.3742,],
    #                     nsteps=500000,
    #                     dt=0.002,
    #                     ens='npt',
    #                     temp=400.0,
    #                     pres = 1.0, 
    #                     tau_t = 0.1,
    #                     tau_p = 0.5,
    #                     prt_freq = 100, 
    #                     copies = None,
    #                     crystal = 'vega', 
    #                     sparam = soft_param,
    #                     switch = 'three-step',
    #                     step = 'deep_on',
    #                     if_meam = False,
    #                     meam_model = None)
    #     ret1 = textwrap.dedent("""\
    #     clear
    #     # --------------------- VARIABLES-------------------------
    #     variable        NSTEPS          equal 500000
    #     variable        THERMO_FREQ     equal 100
    #     variable        DUMP_FREQ       equal 100
    #     variable        TEMP            equal 400.000000
    #     variable        PRES            equal 1.000000
    #     variable        TAU_T           equal 0.100000
    #     variable        TAU_P           equal 0.500000
    #     variable        LAMBDA          equal 7.5000000000e-02
    #     variable        INV_LAMBDA      equal 9.2500000000e-01
    #     # ---------------------- INITIALIZAITION ------------------
    #     units           metal
    #     boundary        p p p
    #     atom_style      atomic
    #     # --------------------- ATOM DEFINITION ------------------
    #     box             tilt large
    #     read_data       conf.lmp
    #     change_box      all triclinic
    #     mass            1 118.710000
    #     # --------------------- FORCE FIELDS ---------------------
    #     variable        EPSILON equal 0.030000
    #     variable        ONE equal 1
    #     pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
    #     pair_coeff      * * deepmd
    #     pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
    #     fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
    #     compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
    #     group           type_1 type 1
    #     fix             l_spring_1 type_1 spring/self 2.3742000000e+00
    #     fix_modify      l_spring_1 energy yes
    #     variable        l_spring equal f_l_spring_1
    #     # --------------------- MD SETTINGS ----------------------
    #     neighbor        1.0 bin
    #     timestep        0.002
    #     thermo          ${THERMO_FREQ}
    #     compute         allmsd all msd
    #     thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
    #     thermo_modify   format 9 %.16e
    #     thermo_modify   format 10 %.16e
    #     # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
    #     fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
    #     # --------------------- INITIALIZE -----------------------
    #     velocity        all create ${TEMP} 7858
    #     group           first id 1
    #     fix             fc first recenter INIT INIT INIT
    #     fix             fm first momentum 1 linear 1 1 1
    #     velocity        first zero linear
    #     # --------------------- RUN ------------------------------
    #     run             ${NSTEPS}
    #     write_data      out.lmp
    #     """)
    #     ret2 = _gen_lammps_input(**input)
    #     self.assertEqual(ret1, ret2)


    # @patch('numpy.random')
    # def test_meam_deep_on(self, patch_random):
    #     patch_random.randint = MagicMock(return_value=7858)
    #     input = dict ( conf_file='conf.lmp', 
    #                     mass_map=[118.71],
    #                     lamb=0.075,
    #                     model=None,
    #                     m_spring_k=[2.3742,],
    #                     nsteps=500000,
    #                     dt=0.002,
    #                     ens='npt',
    #                     temp=400.0,
    #                     pres = 1.0, 
    #                     tau_t = 0.1,
    #                     tau_p = 0.5,
    #                     prt_freq = 100, 
    #                     copies = None,
    #                     crystal = 'vega', 
    #                     sparam = soft_param,
    #                     switch = 'three-step',
    #                     step = 'deep_on',
    #                     if_meam = True,
    #                     meam_model = meam_model)
    #     ret1 = textwrap.dedent("""\
    #     clear
    #     # --------------------- VARIABLES-------------------------
    #     variable        NSTEPS          equal 500000
    #     variable        THERMO_FREQ     equal 100
    #     variable        DUMP_FREQ       equal 100
    #     variable        TEMP            equal 400.000000
    #     variable        PRES            equal 1.000000
    #     variable        TAU_T           equal 0.100000
    #     variable        TAU_P           equal 0.500000
    #     variable        LAMBDA          equal 7.5000000000e-02
    #     variable        INV_LAMBDA      equal 9.2500000000e-01
    #     # ---------------------- INITIALIZAITION ------------------
    #     units           metal
    #     boundary        p p p
    #     atom_style      atomic
    #     # --------------------- ATOM DEFINITION ------------------
    #     box             tilt large
    #     read_data       conf.lmp
    #     change_box      all triclinic
    #     mass            1 118.710000
    #     # --------------------- FORCE FIELDS ---------------------
    #     variable        EPSILON equal 0.030000
    #     variable        ONE equal 1
    #     pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
    #     pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
    #     pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
    #     fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA
    #     compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE
    #     group           type_1 type 1
    #     fix             l_spring_1 type_1 spring/self 2.3742000000e+00
    #     fix_modify      l_spring_1 energy yes
    #     variable        l_spring equal f_l_spring_1
    #     # --------------------- MD SETTINGS ----------------------
    #     neighbor        1.0 bin
    #     timestep        0.002
    #     thermo          ${THERMO_FREQ}
    #     compute         allmsd all msd
    #     thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
    #     thermo_modify   format 9 %.16e
    #     thermo_modify   format 10 %.16e
    #     # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
    #     fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
    #     # --------------------- INITIALIZE -----------------------
    #     velocity        all create ${TEMP} 7858
    #     group           first id 1
    #     fix             fc first recenter INIT INIT INIT
    #     fix             fm first momentum 1 linear 1 1 1
    #     velocity        first zero linear
    #     # --------------------- RUN ------------------------------
    #     run             ${NSTEPS}
    #     write_data      out.lmp
    #     """)
    #     ret2 = _gen_lammps_input(**input)
    #     # print(ret2)
    #     self.assertEqual(ret1, ret2)

    # @patch('numpy.random')
    # def test_meam_spring_off(self, patch_random):
    #     patch_random.randint = MagicMock(return_value=7858)
    #     input = dict ( conf_file='conf.lmp', 
    #                     mass_map=[118.71],
    #                     lamb=0.075,
    #                     model=None,
    #                     m_spring_k=[2.3742,],
    #                     nsteps=500000,
    #                     dt=0.002,
    #                     ens='npt',
    #                     temp=400.0,
    #                     pres = 1.0, 
    #                     tau_t = 0.1,
    #                     tau_p = 0.5,
    #                     prt_freq = 100, 
    #                     copies = None,
    #                     crystal = 'vega', 
    #                     sparam = soft_param,
    #                     switch = 'three-step',
    #                     step = 'spring_off',
    #                     if_meam = True,
    #                     meam_model = meam_model)
    #     ret1 = textwrap.dedent("""\
    #     clear
    #     # --------------------- VARIABLES-------------------------
    #     variable        NSTEPS          equal 500000
    #     variable        THERMO_FREQ     equal 100
    #     variable        DUMP_FREQ       equal 100
    #     variable        TEMP            equal 400.000000
    #     variable        PRES            equal 1.000000
    #     variable        TAU_T           equal 0.100000
    #     variable        TAU_P           equal 0.500000
    #     variable        LAMBDA          equal 7.5000000000e-02
    #     variable        INV_LAMBDA      equal 9.2500000000e-01
    #     # ---------------------- INITIALIZAITION ------------------
    #     units           metal
    #     boundary        p p p
    #     atom_style      atomic
    #     # --------------------- ATOM DEFINITION ------------------
    #     box             tilt large
    #     read_data       conf.lmp
    #     change_box      all triclinic
    #     mass            1 118.710000
    #     # --------------------- FORCE FIELDS ---------------------
    #     variable        EPSILON equal 0.030000
    #     variable        INV_EPSILON equal -${EPSILON}
    #     pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
    #     pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
    #     pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
    #     fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
    #     compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
    #     group           type_1 type 1
    #     fix             l_spring_1 type_1 spring/self 2.1961350000e+00
    #     fix_modify      l_spring_1 energy yes
    #     variable        l_spring equal f_l_spring_1
    #     # --------------------- MD SETTINGS ----------------------
    #     neighbor        1.0 bin
    #     timestep        0.002
    #     thermo          ${THERMO_FREQ}
    #     compute         allmsd all msd
    #     thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
    #     thermo_modify   format 9 %.16e
    #     thermo_modify   format 10 %.16e
    #     # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
    #     fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
    #     # --------------------- INITIALIZE -----------------------
    #     velocity        all create ${TEMP} 7858
    #     group           first id 1
    #     fix             fc first recenter INIT INIT INIT
    #     fix             fm first momentum 1 linear 1 1 1
    #     velocity        first zero linear
    #     # --------------------- RUN ------------------------------
    #     run             ${NSTEPS}
    #     write_data      out.lmp
    #     """)
    #     ret2 = _gen_lammps_input(**input)
    #     # print(ret2)
    #     self.assertEqual(ret1, ret2)

    # def test_raise_err(self):
    #     input = dict ( conf_file='conf.lmp', 
    #                     mass_map=[118.71],
    #                     lamb=0.075,
    #                     model="graph.pb",
    #                     m_spring_k=[2.3742,],
    #                     nsteps=500000,
    #                     dt=0.002,
    #                     ens='npt',
    #                     temp=400.0,
    #                     pres = 1.0, 
    #                     tau_t = 0.1,
    #                     tau_p = 0.5,
    #                     prt_freq = 100, 
    #                     copies = None,
    #                     crystal = 'vega', 
    #                     sparam = soft_param,
    #                     switch = 'three-step',
    #                     step = 'spring_off',
    #                     if_meam = False,
    #                     meam_model = None)
        
    #     input2 = input.copy()
    #     input2['step'] = 'foo'
    #     with self.assertRaises(RuntimeError):
    #         _gen_lammps_input(**input2)

        
    #     input3 = input.copy()
    #     input3['switch'] = 'bar'
    #     with self.assertRaises(RuntimeError):
    #         _gen_lammps_input(**input3)

    #     input4 = input.copy()
    #     input4['ens'] = 'baz'
    #     with self.assertRaises(RuntimeError):
    #         _gen_lammps_input(**input4)




import os, textwrap
import numpy as np
import unittest
from context import dpti
from potential_common import soft_param, soft_param_three_element, meam_model

class TestHtiFfSpring(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_spring_not_var_spring(self):
        input = dict(lamb=0.075, m_spring_k=[118.71],
            var_spring=False)
        ret1 = textwrap.dedent("""\
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_spring(**input)
        self.assertEqual(ret1, ret2)

    def test_spring_var_spring(self):
        input = dict(lamb=0.075, m_spring_k=[118.71],
            var_spring=True)
        ret1 = textwrap.dedent("""\
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.0980675000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_spring(**input)
        self.assertEqual(ret1, ret2)

    def test_spring_multiple_element(self):
        input = dict(lamb=0.075, m_spring_k=[118.71, 207.2],
            var_spring=False)
        ret1 = textwrap.dedent("""\
        group           type_1 type 1
        group           type_2 type 2
        fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        fix_modify      l_spring_1 energy yes
        fix             l_spring_2 type_2 spring/self 2.0720000000e+02
        fix_modify      l_spring_2 energy yes
        variable        l_spring equal f_l_spring_1+f_l_spring_2
        """)
        ret2 = dpti.hti._ff_spring(**input)
        self.assertEqual(ret1, ret2)

def test_spring_var_spring_multiple_element(self):
        input = dict(lamb=0.20, m_spring_k=[118.71, 207.2],
            var_spring=False)
        ret1 = textwrap.dedent("""\
        group           type_1 type 1
        group           type_2 type 2
        fix             l_spring_1 type_1 spring/self 9.4968000000e+01
        fix_modify      l_spring_1 energy yes
        fix             l_spring_2 type_2 spring/self 1.6576000000e+02
        fix_modify      l_spring_2 energy yes
        variable        l_spring equal f_l_spring_1+f_l_spring_2
        """)
        ret2 = dpti.hti._ff_spring(**input)
        self.assertEqual(ret1, ret2)


if __name__ == '__main__':
    unittest.main()



import os, textwrap
import numpy as np
import unittest
import shutil
from context import dpti
from dpti.lib.utils import parse_seq, block_avg, relative_link_file
from dpti.lib.utils import integrate_simpson, integrate_range_hti
from numpy.testing import assert_almost_equal

lambda_seq = [
    "0.00:0.05:0.010",
    "0.05:0.15:0.020",
    "0.15:0.35:0.040",
    "0.35:1.00:0.065",
    "1"
]

class TestParseSeq(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_normal(self):
        array1 = parse_seq(lambda_seq)
        array2 = np.asarray([0.0, 0.01, 0.02, 
            0.03, 0.04, 0.05, 
            0.07, 0.09, 0.11, 
            0.13, 0.15, 0.19, 
            0.23, 0.27, 0.31, 
            0.35, 0.415, 0.48, 
            0.545, 0.61, 0.675, 
            0.74, 0.805, 0.87, 
            0.935, 1.0])
        assert_almost_equal(array1, array2, decimal=10)

    def test_protect_eps(self):
        array1 = parse_seq(lambda_seq, protect_eps=1e-6)
        array2 = np.asarray([0.000001, 0.01, 0.02, 
            0.03, 0.04, 0.05, 
            0.07, 0.09, 0.11, 
            0.13, 0.15, 0.19, 
            0.23, 0.27, 0.31, 
            0.35, 0.415, 0.48, 
            0.545, 0.61, 0.675, 
            0.74, 0.805, 0.87, 
            0.935, 0.999999])
        assert_almost_equal(array1, array2, decimal=10)

    def test_no_posi_args(self):
        with self.assertRaises(TypeError):
            parse_seq(lambda_seq, 1e-6)
            # assert_almost_equal(array1, array2, decimal=10)

class TestBlockAvg(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_normal(self):
        avg1 = 7.158014302
        err1 = 0.262593991
        data_file = 'lammps_test_files/get_thermo.data'
        data_array = np.loadtxt(data_file)
        avg2, err2 = block_avg(data_array[:,1])
        self.assertAlmostEqual(avg1, avg2, places=8)
        self.assertAlmostEqual(err1, err2, places=8)


class TestIntegrateRangeHti(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        # lamb

    def test_lamb_array_odd(self):
        result1 = -0.07001571298782591
        stt_err1 = 2.1394708051996743e-05
        sys_err2 = 2.0797767427780528e-07
        data = np.loadtxt('hti_test_files/odd.hti.out')
        lamb_array = data[:,0]
        dU_array = data[:,1]
        dU_err_array = data[:,2]
        result2, stt_err2, sys_err2 = integrate_range_hti(lamb_array, dU_array, dU_err_array)
        self.assertAlmostEqual(result1, result2, places=8)
        self.assertAlmostEqual(stt_err1, stt_err2, places=8)
        self.assertAlmostEqual(sys_err2, sys_err2, places=8)

    def test_lamb_array_even(self):
        result1 = -35.48046669098458
        stt_err1 = 0.0001625198805022198
        sys_err2 = 8.812949063852216e-07
        data = np.loadtxt('hti_test_files/even.hti.out')
        lamb_array = data[:,0]
        dU_array = data[:,1]
        dU_err_array = data[:,2]
        result2, stt_err2, sys_err2 = integrate_range_hti(lamb_array, dU_array, dU_err_array)
        self.assertAlmostEqual(result1, result2, places=8)
        self.assertAlmostEqual(stt_err1, stt_err2, places=8)
        self.assertAlmostEqual(sys_err2, sys_err2, places=8)

class TestRelativeLinkFile(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        os.mkdir('relative_link_file_test_dir/')
        pass

    def test_normal(self):
        relative_link_file('graph.pb', 'relative_link_file_test_dir/')

    def test_other_place(self):
        relative_link_file('../setup.py', 'relative_link_file_test_dir/')

    def test_abs_path(self):
        abs_path = os.path.abspath(__file__)
        relative_link_file(abs_path, 'relative_link_file_test_dir/')

    def test_abs_path_2(self):
        abs_path = os.path.abspath('../README.md')
        relative_link_file(abs_path, 'relative_link_file_test_dir/')

    def test_raise_err(self):
        with self.assertRaises(RuntimeError):
            relative_link_file('../dpti/', 
            'relative_link_file_test_dir/')

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('relative_link_file_test_dir/')




if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os, sys, json, argparse, glob, shutil
from re import T
import numpy as np
import scipy.constants as pc
import pymbar

# sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../'))
from dpti.einstein import free_energy, frenkel
from dpti.lib.utils import create_path, relative_link_file
from dpti.lib.utils import copy_file_list
from dpti.lib.utils import block_avg
from dpti.lib.utils import integrate_range
from dpti.lib.utils import integrate_range_hti
# from lib.utils import integrate_sys_err
from dpti.lib.utils import compute_nrefine
from dpti.lib.utils import parse_seq
from dpti.lib.utils import get_task_file_abspath
from dpti.lib.utils import get_first_matched_key_from_dict
from dpti.lib.lammps import get_thermo
from dpti.lib.lammps import get_natoms

def make_iter_name (iter_index) :
    return "task_hti." + ('%04d' % iter_index)


def _ff_lj_on(lamb,
              model,
              sparam):
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    # sigma_oo = sparam['sigma_oo']
    # sigma_oh = sparam['sigma_oh']
    # sigma_hh = sparam['sigma_hh']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'pair_style      lj/cut/soft %f %f %f\n' % (nn, alpha_lj, rcut)

    element_num=sparam.get('element_num', 1)
    sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
    for (i, j) in sigma_key_index:
        ret += 'pair_coeff      %s %s ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * ${EPSILON} %f %f\n' % (sigma, activation)
    # ret += 'pair_coeff      1 1 ${EPSILON} %f %f\n' % (sigma_oo, activation)
    # ret += 'pair_coeff      1 2 ${EPSILON} %f %f\n' % (sigma_oh, activation)
    # ret += 'pair_coeff      2 2 ${EPSILON} %f %f\n' % (sigma_hh, activation)
    ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes\n'
    ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON\n'    
    return ret


def _ff_deep_on(lamb,
                model,
                sparam,
                if_meam=False,
                meam_model=None):
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    # sigma_oo = sparam['sigma_oo']
    # sigma_oh = sparam['sigma_oh']
    # sigma_hh = sparam['sigma_hh']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'variable        ONE equal 1\n'
    # if if_meam:
    #     ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f  \n' % (nn, alpha_lj, rcut)
    #     ret += 'pair_coeff      * * meam /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn \n'
    if if_meam:
        ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f\n' % (nn, alpha_lj, rcut)
        ret += f'pair_coeff      * * meam {meam_model["library"]} {meam_model["element"]} {meam_model["potential"]} {meam_model["element"]}\n'
    else:
        ret += 'pair_style      hybrid/overlay deepmd %s lj/cut/soft %f %f %f\n' % (model, nn, alpha_lj, rcut)
        ret += 'pair_coeff      * * deepmd\n'

    element_num=sparam.get('element_num', 1)
    sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
    for (i, j) in sigma_key_index:
        ret += 'pair_coeff      %s %s lj/cut/soft ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * lj/cut/soft ${EPSILON} %f %f\n' % (sigma, activation)
    # ret += 'pair_coeff      1 1 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oo, activation)
    # ret += 'pair_coeff      1 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oh, activation)
    # ret += 'pair_coeff      2 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_hh, activation)
    if if_meam:
        ret += 'fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA\n'
        ret += 'compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE\n'
    else:
        ret += 'fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA\n'
        ret += 'compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE\n'
    return ret

# def _ff_meam_on(lamb,
#                 model,
#                 sparam):
#     nn = sparam['n']
#     alpha_lj = sparam['alpha_lj']
#     rcut = sparam['rcut']
#     epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    # sigma_oo = sparam['sigma_oo']
    # sigma_oh = sparam['sigma_oh']
    # sigma_hh = sparam['sigma_hh']
#     activation = sparam['activation']
#     ret = ''
#     ret += 'variable        EPSILON equal %f\n' % epsilon
#     ret += 'variable        ONE equal 1\n'
#     ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f  \n' % (nn, alpha_lj, rcut)
#     ret += 'pair_coeff      * * meam /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn \n'

#     element_num=sparam.get('element_num', 1)
#     sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
#     for (i, j) in sigma_key_index:
#         ret += 'pair_coeff      %s %s lj/cut/soft ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * lj/cut/soft ${EPSILON} %f %f\n' % (sigma, activation)
    # ret += 'pair_coeff      1 1 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oo, activation)
    # ret += 'pair_coeff      1 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oh, activation)
    # ret += 'pair_coeff      2 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_hh, activation)
#     if if_meam:
#         ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f  \n' % (nn, alpha_lj, rcut)
#         ret += 'pair_coeff      * * meam /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn \n'
#     else:
#         ret += 'fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA\n'
#         ret += 'compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE\n'
#     return ret


def _ff_lj_off(lamb,
               model, 
               sparam,
               if_meam=False,
               meam_model=None) :
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    # sigma_oo = sparam['sigma_oo']
    # sigma_oh = sparam['sigma_oh']
    # sigma_hh = sparam['sigma_hh']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'variable        INV_EPSILON equal -${EPSILON}\n'
    # if if_meam:
    #     ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f  \n'  % (nn, alpha_lj, rcut)
    #     ret += 'pair_coeff      * * meam /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn\n'
    if if_meam:
        ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f\n'  % (nn, alpha_lj, rcut)
        ret += f'pair_coeff      * * meam {meam_model["library"]} {meam_model["element"]} {meam_model["potential"]} {meam_model["element"]}\n'
        # ret += f'pair_coeff      * * meam {meam_model[0]} {meam_model[2]} {meam_model[1]} {meam_model[2]}\n'
    else:
        ret += 'pair_style      hybrid/overlay deepmd %s lj/cut/soft %f %f %f\n' % (model, nn, alpha_lj, rcut)
        ret += 'pair_coeff      * * deepmd\n'
        

    element_num=sparam.get('element_num', 1)
    sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
    for (i, j) in sigma_key_index:
        ret += 'pair_coeff      %s %s lj/cut/soft ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * lj/cut/soft ${EPSILON} %f %f\n' % (sigma, activation)
    # ret += 'pair_coeff      1 1 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oo, activation)
    # ret += 'pair_coeff      1 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oh, activation)
    # ret += 'pair_coeff      2 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_hh, activation)
    ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes\n'
    ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON\n'    
    return ret

# def _ff_meam_lj_off(lamb,
#                model, 
#                sparam) :
#     nn = sparam['n']
#     alpha_lj = sparam['alpha_lj']
#     rcut = sparam['rcut']
#     epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    # sigma_oo = sparam['sigma_oo']
    # sigma_oh = sparam['sigma_oh']
    # sigma_hh = sparam['sigma_hh']
#     activation = sparam['activation']
#     ret = ''
#     ret += 'variable        EPSILON equal %f\n' % epsilon
#     ret += 'variable        INV_EPSILON equal -${EPSILON}\n'
#     ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f  \n'  % (nn, alpha_lj, rcut)
#     ret += 'pair_coeff      * * meam /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn\n'

#     element_num=sparam.get('element_num', 1)
#     sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
#     for (i, j) in sigma_key_index:
#         ret += 'pair_coeff      %s %s lj/cut/soft ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * lj/cut/soft ${EPSILON} %f %f\n' % (sigma, activation)
    # ret += 'pair_coeff      1 1 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oo, activation)
    # ret += 'pair_coeff      1 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oh, activation)
    # ret += 'pair_coeff      2 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_hh, activation)
#     ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes\n'
#     ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON\n'    
#     return ret
    
    

def _ff_spring(lamb,
               m_spring_k,
               var_spring):
    ret = ''
    ntypes = len(m_spring_k)
    for ii in range(ntypes) :
        ret += 'group           type_%s type %s\n' % (ii+1, ii+1)
    for ii in range(ntypes) :
        if var_spring:
            m_spring_const = m_spring_k[ii] * (1 - lamb)
        else:
            m_spring_const = m_spring_k[ii]
        ret += 'fix             l_spring_%s type_%s spring/self %.10e\n' % (ii+1, ii+1, m_spring_const)
        ret += 'fix_modify      l_spring_%s energy yes\n' % (ii+1)
    sum_str = 'f_l_spring_1'
    for ii in range(1,ntypes) :
        sum_str += '+f_l_spring_%s' % (ii+1)
    ret += 'variable        l_spring equal %s\n' % (sum_str)
    return ret

    
def _ff_soft_lj(lamb,
                model,
                m_spring_k,
                step,
                sparam,
                if_meam=False,
                meam_model=None):
    ret = ''
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    if step == 'lj_on':
        ret += _ff_lj_on(lamb, model, sparam)
        var_spring = False
    elif step == 'deep_on':
        # ret += _ff_meam_on(lamb, model, sparam)
        ret += _ff_deep_on(lamb, model, sparam, if_meam=if_meam, meam_model=meam_model)
        var_spring = False
    elif step == 'spring_off':
        # ret += _ff_meam_lj_off(lamb, model, sparam)
        ret += _ff_lj_off(lamb, model, sparam, if_meam=if_meam, meam_model=meam_model)
        var_spring = True
    else:
        raise RuntimeError('unkown step', step)

    ret += _ff_spring(lamb, m_spring_k, var_spring)
    
    return ret


def _ff_two_steps(lamb,
                  model,
                  m_spring_k,
                  step):
    ret = ''
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    ret += 'pair_style      deepmd %s\n' % model
    ret += 'pair_coeff\n'
    
    if step == 'both' or step == 'spring_off':                
        var_spring = True
    elif step == 'deep_on':
        var_spring = False
    else:
        raise RuntimeError('unkown step', step)
    if step == 'both' or step == 'deep_on':                
        var_deep = True
    elif step == 'spring_off':
        var_deep = False
    else:
        raise RuntimeError('unkown step', step)

    ret += _ff_spring(lamb, m_spring_k, var_spring)

    if var_deep:
        ret += 'fix             l_deep all adapt 1 pair deepmd scale * * v_LAMBDA\n'
    ret += 'compute         e_deep all pe pair\n'
    return ret

    
def _gen_lammps_input (conf_file, 
                       mass_map,
                       lamb,
                       model,
                       m_spring_k,
                       nsteps,
                       timestep,
                       ens,
                       temp,
                       pres = 1.0, 
                       tau_t = 0.1,
                       tau_p = 0.5,
                       thermo_freq = 100, 
                       copies = None,
                       crystal = 'vega', 
                       sparam = {},
                       switch = 'one-step',
                       step = 'both',
                       if_meam = False,
                       meam_model = None):
    ret = ''
    ret += 'clear\n'
    ret += '# --------------------- VARIABLES-------------------------\n'
    ret += 'variable        NSTEPS          equal %d\n' % nsteps
    ret += 'variable        THERMO_FREQ     equal %d\n' % thermo_freq
    ret += 'variable        DUMP_FREQ       equal %d\n' % thermo_freq
    ret += 'variable        TEMP            equal %f\n' % temp
    ret += 'variable        PRES            equal %f\n' % pres
    ret += 'variable        TAU_T           equal %f\n' % tau_t
    ret += 'variable        TAU_P           equal %f\n' % tau_p
    ret += 'variable        LAMBDA          equal %.10e\n' % lamb
    ret += 'variable        INV_LAMBDA      equal %.10e\n' % (1-lamb)
    ret += '# ---------------------- INITIALIZAITION ------------------\n'
    ret += 'units           metal\n'
    ret += 'boundary        p p p\n'
    ret += 'atom_style      atomic\n'
    ret += '# --------------------- ATOM DEFINITION ------------------\n'
    ret += 'box             tilt large\n'
    ret += 'read_data       %s\n' % conf_file
    if copies is not None :
        ret += 'replicate       %d %d %d\n' % (copies[0], copies[1], copies[2])
    ret += 'change_box      all triclinic\n'
    for jj in range(len(mass_map)) :
        ret += "mass            %d %f\n" %(jj+1, mass_map[jj])

    # force field setting
    if switch == 'one-step' or switch == 'two-step':
        ret += _ff_two_steps(lamb, model, m_spring_k, step)
    elif switch == 'three-step':
        ret += _ff_soft_lj(lamb, model, m_spring_k, step, sparam, if_meam=if_meam, meam_model=meam_model)
    else:
        raise RuntimeError('unknow switch', switch)

    ret += '# --------------------- MD SETTINGS ----------------------\n'    
    ret += 'neighbor        1.0 bin\n'
    ret += 'timestep        %s\n' % timestep
    ret += 'thermo          ${THERMO_FREQ}\n'
    ret += 'compute         allmsd all msd\n'
    if 1 - lamb != 0 :
        if type(m_spring_k) is not list :        
            if switch == 'three-step':
                ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol f_l_spring c_e_diff[1] c_allmsd[*]\n'
            else:
                ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol f_l_spring c_e_deep c_allmsd[*]\n'
        else :
            if switch == 'three-step':
                ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]\n'
            else:
                ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_deep c_allmsd[*]\n'
    else :
        ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol c_e_deep c_e_deep c_allmsd[*]\n'
    ret += 'thermo_modify   format 9 %.16e\n'
    ret += 'thermo_modify   format 10 %.16e\n'
    ret += '# dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz\n'
    if ens == 'nvt' :
        ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
    elif ens == 'nvt-langevin' :
        ret += 'fix             1 all nve\n'
        ret += 'fix             2 all langevin ${TEMP} ${TEMP} ${TAU_T} %d' % (np.random.randint(1, 2**16))
        if crystal == 'frenkel':
            ret += ' zero yes\n'
        else:
            ret += ' zero no\n'            
    elif ens == 'npt-iso' or ens == 'npt':
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'nve' :
        ret += 'fix             1 all nve\n'
    else :
        raise RuntimeError('unknow ensemble %s\n' % ens)        
    ret += '# --------------------- INITIALIZE -----------------------\n'    
    ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
    if crystal == 'frenkel' :
        ret += 'fix             fc all recenter INIT INIT INIT\n'
        ret += 'fix             fm all momentum 1 linear 1 1 1\n'
        ret += 'velocity        all zero linear\n'
    elif crystal == 'vega' :
        ret += 'group           first id 1\n'
        ret += 'fix             fc first recenter INIT INIT INIT\n'
        ret += 'fix             fm first momentum 1 linear 1 1 1\n'
        ret += 'velocity        first zero linear\n'
    else :
        raise RuntimeError('unknow crystal ' + crystal)
    ret += '# --------------------- RUN ------------------------------\n'    
    ret += 'run             ${NSTEPS}\n'
    ret += 'write_data      out.lmp\n'
    
    return ret

# def _gen_lammps_input_ideal (conf_file, 
#                              mass_map,
#                              lamb,
#                              model,
#                              nsteps,
#                              dt,
#                              ens,
#                              temp,
#                              pres = 1.0, 
#                              tau_t = 0.1,
#                              tau_p = 0.5,
#                              prt_freq = 100, 
#                              copies = None,
#                              norm_style = 'first',
#                              if_meam = False,
#                              meam_model = None) :
#     ret = ''
#     ret += 'clear\n'
#     ret += '# --------------------- VARIABLES-------------------------\n'
#     ret += 'variable        NSTEPS          equal %d\n' % nsteps
#     ret += 'variable        THERMO_FREQ     equal %d\n' % prt_freq
#     ret += 'variable        DUMP_FREQ       equal %d\n' % prt_freq
#     ret += 'variable        TEMP            equal %f\n' % temp
#     ret += 'variable        PRES            equal %f\n' % pres
#     ret += 'variable        TAU_T           equal %f\n' % tau_t
#     ret += 'variable        TAU_P           equal %f\n' % tau_p
#     ret += 'variable        LAMBDA          equal %.10e\n' % lamb
#     ret += 'variable        ZERO            equal 0\n'
#     ret += '# ---------------------- INITIALIZAITION ------------------\n'
#     ret += 'units           metal\n'
#     ret += 'boundary        p p p\n'
#     ret += 'atom_style      atomic\n'
#     ret += '# --------------------- ATOM DEFINITION ------------------\n'
#     ret += 'box             tilt large\n'
#     ret += 'read_data       %s\n' % conf_file
#     if copies is not None :
#         ret += 'replicate       %d %d %d\n' % (copies[0], copies[1], copies[2])
#     ret += 'change_box      all triclinic\n'
#     for jj in range(len(mass_map)) :
#         ret += "mass            %d %f\n" %(jj+1, mass_map[jj])
#     ret += '# --------------------- FORCE FIELDS ---------------------\n'
#     ret += 'pair_style      deepmd %s\n' % model
#     ret += 'pair_coeff\n'
#     ret += 'fix             l_deep all adapt 1 pair deepmd scale * * v_LAMBDA\n'
#     ret += 'compute         e_deep all pe pair\n'
#     ret += '# --------------------- MD SETTINGS ----------------------\n'    
#     ret += 'neighbor        1.0 bin\n'
#     ret += 'timestep        %s\n' % dt
#     ret += 'thermo          ${THERMO_FREQ}\n'
#     ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol v_ZERO c_e_deep c_allmsd[*]\n'
#     ret += 'thermo_modify   format 10 %.16e\n'
#     ret += '# dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz\n'
#     if ens == 'nvt' :
#         ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
#     elif ens == 'nvt-langevin' :
#         ret += 'fix             1 all nve\n'
#         ret += 'fix             2 all langevin ${TEMP} ${TEMP} ${TAU_T} %d zero yes\n' % (np.random.randint(1, 2**16))
#     elif ens == 'npt-iso' or ens == 'npt':
#         ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
#     elif ens == 'nve' :
#         ret += 'fix             1 all nve\n'
#     else :
#         raise RuntimeError('unknow ensemble %s\n' % ens)        
#     ret += 'fix             mzero all momentum 10 linear 1 1 1\n'
#     ret += '# --------------------- INITIALIZE -----------------------\n'    
#     ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
#     ret += 'velocity        all zero linear\n'
#     ret += '# --------------------- RUN ------------------------------\n'    
#     ret += 'run             ${NSTEPS}\n'
#     ret += 'write_data      out.lmp\n'
    
#     return ret


def make_tasks(iter_name, jdata, ref='einstein', switch = 'one-step', if_meam=None):
    if if_meam is None:
        if_meam = jdata.get('if_meam', False)
    equi_conf = os.path.abspath(jdata['equi_conf'])
    meam_model = jdata.get('meam_model', None)
    model = os.path.abspath(jdata['model'])

    if if_meam is None:
        if_meam = jdata.get('if_meam', None)

    if switch == 'one-step':
        subtask_name = iter_name
        _make_tasks(subtask_name, jdata, ref, step = 'both', if_meam=if_meam,  meam_model=meam_model)
        if if_meam:
            relative_link_file(meam_model['library'], iter_name)
            relative_link_file(meam_model['potential'], iter_name)
        else:
            pass
    elif switch == 'two-step' or switch == 'three-step':
        job_abs_dir = create_path(iter_name)
        copied_conf = os.path.join(os.path.abspath(iter_name), 'conf.lmp')
        shutil.copyfile(equi_conf, copied_conf)
        jdata['equi_conf'] = 'conf.lmp'
        linked_model = os.path.join(os.path.abspath(iter_name), 'graph.pb')

        if if_meam:
            relative_link_file(meam_model['library'], job_abs_dir)
            relative_link_file(meam_model['potential'], job_abs_dir)
        else:
            pass

        shutil.copyfile(model, linked_model)
        jdata['model'] = 'graph.pb'
        cwd = os.getcwd()
        os.chdir(iter_name)    
        with open('in.json', 'w') as fp:
            json.dump(jdata, fp, indent=4)
        if switch == 'two-step':
            subtask_name = '00.deep_on'
            _make_tasks(subtask_name, jdata, ref, switch = switch, step = 'deep_on', link = True, if_meam=if_meam, meam_model=meam_model)
            subtask_name = '01.spring_off'
            _make_tasks(subtask_name, jdata, ref, switch = switch, step = 'spring_off', link = True, if_meam=if_meam, meam_model=meam_model)
        elif switch == 'three-step':
            subtask_name = '00.lj_on'
            _make_tasks(subtask_name, jdata, ref, switch = switch, step = 'lj_on', link = True, if_meam=if_meam, meam_model=meam_model)
            subtask_name = '01.deep_on'
            _make_tasks(subtask_name, jdata, ref, switch = switch, step = 'deep_on', link = True, if_meam=if_meam, meam_model=meam_model)
            subtask_name = '02.spring_off'
            _make_tasks(subtask_name, jdata, ref, switch = switch, step = 'spring_off', link = True, if_meam=if_meam,  meam_model=meam_model)
        else:
            raise RuntimeError('unknow switch', switch)
        os.chdir(cwd)
    else:
        raise RuntimeError('unknow switch', switch)

    
def _make_tasks(iter_name, jdata, ref, switch = 'one-step', step = 'both', link = False, if_meam=False, meam_model=None):
    if 'crystal' not in jdata:
        print('do not find crystal in jdata, assume vega')
        jdata['crystal'] = 'vega'

    crystal = jdata['crystal']
    protect_eps = jdata['protect_eps']

    if switch == 'one-step':
        all_lambda = parse_seq(jdata['lambda'])
    elif switch == 'two-step' or switch == 'three-step':
        if step == 'deep_on':
            all_lambda = parse_seq(jdata['lambda_deep_on'])
        elif step == 'spring_off':
            all_lambda = parse_seq(jdata['lambda_spring_off'])
        elif step == 'lj_on':
            all_lambda = parse_seq(jdata['lambda_lj_on'])
        else:
            raise RuntimeError('unknown step', step)

    if all_lambda[0] == 0 :
        all_lambda[0] += protect_eps
    if all_lambda[-1] == 1 :
        all_lambda[-1] -= protect_eps
        
    equi_conf = jdata['equi_conf']
    equi_conf = os.path.abspath(equi_conf)
    model = jdata['model']
    model = os.path.abspath(model)
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    nsteps = jdata['nsteps']
    # timestep = jdata['timestep']
    timestep = get_first_matched_key_from_dict(jdata, ['timestep', 'dt'])
    spring_k = jdata['spring_k']

    sparam = jdata.get('soft_param', {})
    if sparam:
        # update for fields in jsons relating to water

        if 'sigma_oo' in sparam:
            sparam['sigma_0_0'] = sparam['sigma_oo']
            sparam['sigma_0_1'] = sparam['sigma_oh']
            sparam['sigma_1_1'] = sparam['sigma_hh']

        element_num = len(mass_map)
        sparam['element_num'] = element_num

        sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
        sigma_key_name_list = ['sigma_'+str(t[0])+'_'+str(t[1]) for t in sigma_key_index ]
        for sigma_key_name in sigma_key_name_list:
            assert sparam.get(sigma_key_name, None), 'there must be key-value for {sigma_key_name} in soft_param'.format(sigma_key_name=sigma_key_name)

    if crystal == 'frenkel' :
        m_spring_k = []
        for ii in mass_map :
            m_spring_k.append(spring_k * ii)
    if crystal == 'vega' :
        m_spring_k = []
        for ii in mass_map :
            m_spring_k.append(spring_k * ii)
    # thermo_freq = jdata['thermo_freq']
    thermo_freq = get_first_matched_key_from_dict(jdata, ['thermo_freq', 'stat_freq'])
    copies = None
    if 'copies' in jdata :
        copies = jdata['copies']
    temp = jdata['temp']
    jdata['reference'] = ref
    jdata['switch'] = switch
    jdata['step'] = step

    create_path(iter_name)
    copied_conf = os.path.join(os.path.abspath(iter_name), 'conf.lmp')
    if not link :
        shutil.copyfile(equi_conf, copied_conf)
    else:
        cwd = os.getcwd()
        os.chdir(iter_name)
        os.symlink(os.path.relpath(equi_conf), 'conf.lmp')
        os.chdir(cwd)
    jdata['equi_conf'] = 'conf.lmp'    
    linked_model = os.path.join(os.path.abspath(iter_name), 'graph.pb')
    if not link:
        shutil.copyfile(model, linked_model)
    else:
        cwd = os.getcwd()
        os.chdir(iter_name)
        os.symlink(os.path.relpath(model), 'graph.pb')
        os.chdir(cwd)
    jdata['model'] = 'graph.pb'
    langevin = jdata.get('langevin', True)

    cwd = os.getcwd()
    os.chdir(iter_name)
    with open('in.json', 'w') as fp:
        json.dump(jdata, fp, indent=4)
    os.chdir(cwd)

    for idx,ii in enumerate(all_lambda) :
        work_path = os.path.join(iter_name, 'task.%06d' % idx)
        create_path(work_path)
        os.chdir(work_path)
        os.symlink(os.path.relpath(copied_conf), 'conf.lmp')
        os.symlink(os.path.relpath(linked_model), 'graph.pb')
        if if_meam:
            meam_library_basename = os.path.basename(meam_model['library'])
            meam_potential_basename = os.path.basename(meam_model['potential'])
            relative_link_file(os.path.join('../../', meam_library_basename), './')
            relative_link_file(os.path.join('../../', meam_potential_basename), './')
        if idx == 0:
            ens = 'nvt-langevin'
        else :
            ens = 'nvt'
        if langevin:
            ens = 'nvt-langevin'        
        if jdata.get('ens', False):
            ens = jdata.get('ens')
        if ref == 'einstein' :
            lmp_str \
                = _gen_lammps_input('conf.lmp',
                                    mass_map, 
                                    ii, 
                                    'graph.pb',
                                    m_spring_k, 
                                    nsteps, 
                                    timestep,
                                    ens,
                                    temp,
                                    thermo_freq = thermo_freq, 
                                    copies = copies,
                                    switch = switch,
                                    step = step,
                                    sparam = sparam,
                                    crystal = crystal,
                                    if_meam = if_meam,
                                    meam_model = meam_model)
        elif ref == 'ideal':
            raise RuntimeError('choose hti_liq.py')
            # lmp_str \
            #     = _gen_lammps_input_ideal('conf.lmp',
            #                               model_mass_map, 
            #                               ii, 
            #                               'graph.pb',
            #                               nsteps, 
            #                               dt,
            #                               ens,
            #                               temp,
            #                               prt_freq = stat_freq, 
            #                               copies = copies,
            #                               if_meam = if_meam,
            #                               meam_model = meam_model)
        else :
            raise RuntimeError('unknow reference system type ' + ref)
        with open('in.lammps', 'w') as fp :
            fp.write(lmp_str)
        with open('lambda.out', 'w') as fp :
            fp.write(str(ii))
        os.chdir(cwd)


def refine_task (from_task, to_task, err, print_ref=False, if_meam=None, meam_model=None) :
    # raise RuntimeError('No entry')
    from_task = os.path.abspath(from_task)
    to_task = os.path.abspath(to_task)
    
    from_ti = os.path.join(from_task, 'hti.out')
    if not os.path.isfile(from_ti) :
        raise RuntimeError("cannot find file %s, task should be computed befor refined" % from_ti)
    tmp_array = np.loadtxt(from_ti)
    all_t = tmp_array[:,0]
    integrand = tmp_array[:,1]
    ntask = all_t.size

    interval_nrefine = compute_nrefine(all_t, integrand, err)
    if print_ref:
        print(interval_nrefine)
        return

    refined_t = []
    back_map = []
    for ii in range(0, ntask-1) :
        refined_t.append(all_t[ii])
        back_map.append(ii)
        hh = (all_t[ii+1] - all_t[ii]) / interval_nrefine[ii]
        for jj in range(1, interval_nrefine[ii]) :
            refined_t.append(all_t[ii] + jj * hh)
            back_map.append(-1)
    refined_t.append(all_t[-1])
    back_map.append(ntask-1)

    from_json = os.path.join(from_task, 'in.json')
    to_json = os.path.join(to_task, 'in.json')
    from_jdata = json.load(open(from_json))
    to_jdata = from_jdata

    to_jdata['lambda'] = refined_t
    to_jdata['orig_task'] = from_task
    to_jdata['back_map'] = back_map
    to_jdata['refine_error'] = err
    to_jdata['equi_conf'] = get_task_file_abspath(from_task, from_jdata['equi_conf'])
    to_jdata['model'] = get_task_file_abspath(from_task, from_jdata['model'])

    make_tasks(to_task, to_jdata, to_jdata['reference'], if_meam=if_meam)

    from_task_list = glob.glob(os.path.join(from_task, 'task.[0-9]*'))
    from_task_list.sort()
    to_task_list = glob.glob(os.path.join(to_task, 'task.[0-9]*'))
    to_task_list.sort()
    assert(len(from_task_list) == ntask)
    assert(len(to_task_list) == len(refined_t))

    for ii in range(len(to_task_list)) :
        if back_map[ii] < 0 : 
            continue
        for jj in ['data', 'log.lammps'] :
            shutil.copyfile(
                os.path.join(from_task_list[back_map[ii]], jj), 
                os.path.join(to_task_list[ii], jj), 
            )
        with open(os.path.join(to_task_list[ii], 'from.dir'), 'w') as fp:
            fp.write(from_task_list[back_map[ii]])
    

def _compute_thermo(fname, natoms, stat_skip, stat_bsize) :
    data = get_thermo(fname)
    ea, ee = block_avg(data[:, 3], skip = stat_skip, block_size = stat_bsize)
    ha, he = block_avg(data[:, 4], skip = stat_skip, block_size = stat_bsize)
    ta, te = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
    pa, pe = block_avg(data[:, 6], skip = stat_skip, block_size = stat_bsize)
    va, ve = block_avg(data[:, 7], skip = stat_skip, block_size = stat_bsize)
    thermo_info = {}
    thermo_info['p'] = pa
    thermo_info['p_err'] = pe
    thermo_info['v'] = va / natoms
    thermo_info['v_err'] = ve / np.sqrt(natoms)
    thermo_info['e'] = ea / natoms
    thermo_info['e_err'] = ee / np.sqrt(natoms)
    thermo_info['h'] = ha / natoms
    thermo_info['h_err'] = he / np.sqrt(natoms)
    thermo_info['t'] = ta
    thermo_info['t_err'] = te
    unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
    thermo_info['pv'] = pa * va * unit_cvt / natoms
    thermo_info['pv_err'] = pe * va * unit_cvt  / np.sqrt(natoms)
    return thermo_info


def post_tasks(iter_name, jdata, natoms = None, method = 'inte', scheme = 's'):
    switch = 'one-step'
    if os.path.isdir(os.path.join(iter_name, '00.deep_on')):
        switch = 'two-step'
    if os.path.isdir(os.path.join(iter_name, '00.lj_on')):
        switch = 'three-step'

    if switch == 'two-step':
        subtask_name = os.path.join(iter_name, '00.deep_on')
        if method == 'inte' :
            e0, err0, tinfo0 = _post_tasks(subtask_name, jdata, natoms = natoms, scheme = scheme, switch = switch, step = 'deep_on')
        elif method == 'mbar':
            e0, err0, tinfo0 = _post_tasks_mbar(subtask_name, jdata, natoms = natoms, switch = switch, step = 'deep_on')
        else :
            raise RuntimeError('unknow method for integration')
        print('# fe of deep_on:    %20.12f  %10.3e %10.3e' % (e0, err0[0], err0[1]))
        subtask_name = os.path.join(iter_name, '01.spring_off')
        if method == 'inte' :
            e1, err1, tinfo1 = _post_tasks(subtask_name, jdata, natoms = natoms, scheme = scheme, switch = switch, step = 'spring_off')
        elif method == 'mbar':
            e1, err1, tinfo1 = _post_tasks_mbar(subtask_name, jdata, natoms = natoms, switch = switch, step = 'spring_off')
        else :
            raise RuntimeError('unknow method for integration')
        print('# fe of spring_off: %20.12f  %10.3e %10.3e' % (e1, err1[0], err1[1]))
        de = e0 + e1
        stt_err = np.sqrt(np.square(err0[0]) + np.square(err1[0]))
        sys_err = ((err0[1]) + (err1[1]))
        err = [stt_err, sys_err]
        tinfo = tinfo1
    elif switch == 'three-step':
        subtask_name = os.path.join(iter_name, '00.lj_on')
        print(f'# HTI three-step integration [value, stt_err, sys_err]')
        if method == 'inte' :
            e0, err0, tinfo0 = _post_tasks(subtask_name, jdata, natoms = natoms, scheme = scheme, switch = switch, step = 'lj_on')
        elif method == 'mbar':
            e0, err0, tinfo0 = _post_tasks_mbar(subtask_name, jdata, natoms = natoms, switch = switch, step = 'lj_on')
        else :
            raise RuntimeError('unknow method for integration')
        print('# fe of lj_on:      %20.12f  %10.3e %10.3e' % (e0, err0[0], err0[1]))
        subtask_name = os.path.join(iter_name, '01.deep_on')
        if method == 'inte' :
            e1, err1, tinfo1 = _post_tasks(subtask_name, jdata, natoms = natoms, scheme = scheme, switch = switch, step = 'deep_on')
        elif method == 'mbar':
            e1, err1, tinfo1 = _post_tasks_mbar(subtask_name, jdata, natoms = natoms, switch = switch, step = 'deep_on')
        else :
            raise RuntimeError('unknow method for integration')
        print('# fe of deep_on:   %20.12f  %10.3e %10.3e' % (e1, err1[0], err1[1]))
        subtask_name = os.path.join(iter_name, '02.spring_off')
        if method == 'inte' :
            e2, err2, tinfo2 = _post_tasks(subtask_name, jdata, natoms = natoms, scheme = scheme, switch = switch, step = 'spring_off')
        elif method == 'mbar':
            e2, err2, tinfo2 = _post_tasks_mbar(subtask_name, jdata, natoms = natoms, switch = switch, step = 'spring_off')
        else :
            raise RuntimeError('unknow method for integration')
        print('# fe of spring_off: %20.12f  %10.3e %10.3e' % (e2, err2[0], err2[1]))
        de = e0 + e1 + e2
        stt_err = np.sqrt(np.square(err0[0]) + np.square(err1[0]) + np.square(err2[0]))
        sys_err = ((err0[1]) + (err1[1]) + (err2[1]))
        err = [stt_err, sys_err]
        tinfo = tinfo2
    else:
        if method == 'inte':
            de, err, tinfo = _post_tasks(iter_name, jdata, natoms = natoms, scheme = scheme)
        elif method == 'mbar':
            de, err, tinfo = _post_tasks_mbar(iter_name, jdata, natoms = natoms)
    return de, err, tinfo
    

def _post_tasks(iter_name, jdata, natoms = None, scheme = 's', switch = 'one-step', step = 'both') :
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)
    equi_conf = get_task_file_abspath(iter_name, jdata['equi_conf'])
    assert(os.path.isfile(equi_conf))
    if natoms == None :
        natoms = get_natoms(equi_conf)
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
    print('# natoms: %d' % natoms)
    
    all_lambda = []
    all_es = []
    all_es_err = []
    all_ed = []
    all_ed_err = []

    all_etot = []
    all_etot_err = []
    all_enthalpy = []
    all_msd_xyz = []

    for ii in all_tasks :
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        sa, se = block_avg(data[:, 8], skip = stat_skip, block_size = stat_bsize)
        da, de = block_avg(data[:, 9], skip = stat_skip, block_size = stat_bsize)
        etot, etot_err = block_avg(data[:, 3], skip = stat_skip, block_size = stat_bsize)
        enthalpy, _ = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
        msd_xyz = data[-1,-1]
        sa /= natoms
        se /= np.sqrt(natoms)
        da /= natoms
        de /= np.sqrt(natoms)
        lmda_name = os.path.join(ii, 'lambda.out')
        ll = float(open(lmda_name).read())
        all_lambda.append(ll)
        all_es.append(sa)
        all_ed.append(da)
        all_es_err.append(se)
        all_ed_err.append(de)
        
        all_etot.append(etot/natoms)
        all_etot_err.append(etot_err)
        all_enthalpy.append(enthalpy)
        all_msd_xyz.append(msd_xyz)

    all_lambda = np.array(all_lambda)
    all_es = np.array(all_es)
    all_ed = np.array(all_ed)
    all_es_err = np.array(all_es_err)
    all_ed_err = np.array(all_ed_err)
    if switch == 'one-step' or switch == 'two-step':
        if step == 'both':
            de = all_ed / all_lambda - all_es / (1 - all_lambda)
            all_err = np.sqrt(np.square(all_ed_err / all_lambda) + np.square(all_es_err / (1 - all_lambda)))
        elif step == 'deep_on':
            de = all_ed / all_lambda
            all_err = all_ed_err / all_lambda
        elif step == 'spring_off':
            de = -all_es / (1 - all_lambda)
            all_err = all_es_err / (1 - all_lambda)
        else:
            raise RuntimeError('unknow step', step)
    elif switch == 'three-step' :
        if step == 'lj_on' or step == 'deep_on':
            de = all_ed
            all_err = all_ed_err
        elif step == 'spring_off':
            de = -all_es / (1 - all_lambda) + all_ed
            all_err = np.sqrt( np.square(all_es_err / (1 - all_lambda)) +
                               np.square(all_ed_err) )
        else:
            raise RuntimeError('unknow step', step)
    else:
        raise RuntimeError('unknow switch', switch)

    all_print = []
    # all_print.append(np.arange(len(all_lambda)))
    all_print.append(all_lambda)
    all_print.append(de)
    all_print.append(all_err)
    all_print.append(all_ed / all_lambda)
    all_print.append(all_es / (1 - all_lambda))
    all_print.append(all_ed_err / all_lambda)
    all_print.append(all_es_err / (1 - all_lambda))
    all_print.append(all_etot)
    # all_print.append(all_etot_err)
    all_print.append(all_es)
    all_print.append(all_enthalpy)
    all_print.append(all_msd_xyz)
    all_print = np.array(all_print)
    np.savetxt(os.path.join(iter_name, 'hti.out'), 
               all_print.T, 
               fmt = '%.8e', 
               header = 'lmbda dU dU_err Ud Us Ud_err Us_err etot spring_eng enthalpy msd_xyz')

    diff_e, err, sys_err = integrate_range_hti(all_lambda, de, all_err, scheme=scheme)
    # new_lambda, i, i_e, s_e = integrate_range(all_lambda, de, all_err, scheme = scheme)
    # if new_lambda[-1] != all_lambda[-1] :
    #     if new_lambda[-1] == all_lambda[-2]:
    #         _, i1, i_e1, s_e1 = integrate_range(all_lambda[-2:], de[-2:], all_err[-2:], scheme='t')
    #         diff_e = i[-1] + i1[-1]
    #         err = np.linalg.norm([s_e[-1], s_e1[-1]])
    #         sys_err = i_e[-1] + i_e1[-1]
    #     else :
    #         raise RuntimeError("lambda does not match!")
    # else:
    #     diff_e = i[-1]
    #     err = s_e[-1]
    #     sys_err = i_e[-1]
    
    # diff_e, err = integrate(all_lambda, de, all_err)
    # sys_err = integrate_sys_err(all_lambda, de)

    path_endpnt = os.path.join(iter_name, 'task.endpnt')
    if os.path.isdir(path_endpnt) :
        print('# Found end point, compute thermo info from it')
        thermo_info = _compute_thermo(os.path.join(path_endpnt, 'log.lammps'),
                                      natoms,
                                      stat_skip, stat_bsize)
    else :
        print('# Not found end point, compute thermo info from the last lambda')
        thermo_info = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'),
                                      natoms,
                                      stat_skip, stat_bsize)

    return diff_e, [err,sys_err], thermo_info


def _post_tasks_mbar(iter_name, jdata, natoms = None, switch = 'one-step', step = 'both') :
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)
    equi_conf = jdata['equi_conf']
    cwd = os.getcwd()
    os.chdir(iter_name)
    assert(os.path.isfile(equi_conf))
    equi_conf = os.path.abspath(equi_conf)
    os.chdir(cwd)
    temp = jdata['temp']
    if natoms == None :
        natoms = get_natoms(equi_conf)
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
    print('# natoms: %d' % natoms)
    
    all_lambda = []
    for ii in all_tasks :
        lmda_name = os.path.join(ii, 'lambda.out')
        ll = float(open(lmda_name).read())
        all_lambda.append(ll)
    all_lambda = np.array(all_lambda)
    nlambda = all_lambda.size

    ukn = np.array([])
    nk = []
    kt_in_ev = pc.Boltzmann * temp / pc.electron_volt
    for idx,ii in enumerate(all_tasks) :
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        this_ed = data[:,9] / kt_in_ev
        this_es = data[:,8] / kt_in_ev
        this_ed = this_ed[stat_skip::1]
        this_es = this_es[stat_skip::1]
        nk.append(this_ed.size)
        if switch == 'one-step' or switch == 'two-step':
            if step == 'both':
                ed = this_ed / all_lambda[idx]
                es = this_es / (1 - all_lambda[idx])
                block_u = []
                for ll in all_lambda :
                    block_u.append(ed * ll + es * (1-ll))
            elif step == 'deep_on':
                ed = this_ed / all_lambda[idx]
                block_u = []
                for ll in all_lambda :
                    block_u.append(ed * ll)
            elif step == 'spring_off':
                es = this_es / (1 - all_lambda[idx])
                block_u = []
                for ll in all_lambda :
                    block_u.append(es * (1-ll))
            else:
                raise RuntimeError('unknown switch_style', switch)
        elif switch == 'three-step':
            if step == 'lj_on' or step == 'deep_on':
                ed = this_ed
                block_u = []
                for ll in all_lambda:
                    block_u.append(ed * ll)
            elif step == 'spring_off':
                ed = this_ed
                es = this_es / (1 - all_lambda[idx])
                block_u = []
                for ll in all_lambda :
                    block_u.append(ed * ll + es * (1-ll))
            else:
                raise RuntimeError('unknow step', step)
        else:
            raise RuntimeError('unknow switch', switch)
                
        block_u = np.reshape(block_u, [nlambda, -1])
        if ukn.size == 0 :
            ukn = block_u 
        else :
            ukn = np.concatenate((ukn, block_u), axis = 1)
    nk = np.array(nk)

    mbar = pymbar.MBAR(ukn, nk, initialize = 'BAR', relative_tolerance = 1e-9)
    #Deltaf_ij, dDeltaf_ij, Theta_ij = mbar.getFreeEnergyDifferences()
    Deltaf_ij, dDeltaf_ij = mbar.getFreeEnergyDifferences()
    Deltaf_ij = Deltaf_ij / natoms
    dDeltaf_ij = dDeltaf_ij / np.sqrt(natoms)

    diff_e = Deltaf_ij[0,-1] * kt_in_ev
    err = dDeltaf_ij[0,-1] * kt_in_ev

    thermo_info = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'), 
                                  natoms,
                                  stat_skip, stat_bsize)

    return diff_e, [err,0], thermo_info


def print_thermo_info(info) :
    ptr = '# thermodynamics (normalized by natoms)\n'
    ptr += '# E (err)  [eV]:  %20.8f %20.8f\n' % (info['e'], info['e_err'])
    ptr += '# H (err)  [eV]:  %20.8f %20.8f\n' % (info['h'], info['h_err'])
    ptr += '# T (err)   [K]:  %20.8f %20.8f\n' % (info['t'], info['t_err'])
    ptr += '# P (err) [bar]:  %20.8f %20.8f\n' % (info['p'], info['p_err'])
    ptr += '# V (err) [A^3]:  %20.8f %20.8f\n' % (info['v'], info['v_err'])
    ptr += '# PV(err)  [eV]:  %20.8f %20.8f' % (info['pv'], info['pv_err'])
    print(ptr)

def compute_task(job, free_energy_type='helmholtz', method='inte', scheme='simpson', manual_pv=None, manual_pv_err=None):
    # print('hti.compute_task', job, jdata, method, scheme, free_energy_type)
    # assert 'reference' in jdata
    # job = args.JOB
    jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))
    if 'reference' not in jdata:
        jdata['reference'] = 'einstein'
    if jdata['crystal'] == 'vega':
        e0 = free_energy(job)
    if jdata['crystal'] == 'frenkel':
        e0 = frenkel(job)
    de, de_err, thermo_info = post_tasks(job, jdata, method=method, scheme=scheme)
    # printing
    print_format = '%20.12f  %10.3e  %10.3e'
    print_thermo_info(thermo_info)

    info = thermo_info.copy()
    
    if jdata['reference'] == 'einstein' :
        print('# free ener of Einstein Mole: %20.8f' % e0)
    else :
        print('# free ener of ideal gas: %20.8f' % e0)            

    print(('# fe contrib due to integration ' + print_format) \
          % (de, de_err[0], de_err[1]))

    pv = None
    pv_err = None
    
    if free_energy_type == 'helmholtz' :
        e1 = e0 + de
        e1_err = de_err[0]
        print('# Helmholtz free ener per atom (stat_err inte_err) [eV]:')
        print(print_format % (e1, de_err[0], de_err[1]))
    elif free_energy_type == 'gibbs' :
        if manual_pv is None:
            pv = thermo_info['pv']
        else: 
            print(f"# use manual_pv={manual_pv}")
            pv = manual_pv
        if manual_pv_err is None:
            pv_err = thermo_info['pv_err']
        else:
            print(f"# use manual_pv_err={manual_pv_err}")
            pv_err = manual_pv_err
        e1 = e0 + de + pv
        e1_err = np.sqrt(de_err[0]**2 + pv_err**2)
        print('# Gibbs free ener per atom (stat_err inte_err) [eV]:')
        print(print_format % (e1, e1_err, de_err[1]))
    else:
        raise RuntimeError('known free energy type')

    info['free_energy_type'] = free_energy_type
    info['e0'] = e0
    info['pv'] = pv
    info['pv_err'] = pv_err
    info['de'] = de
    info['de_err'] = de_err
    info['e1'] = e1
    info['e1_err'] = e1_err
    with open(os.path.join(job, 'result.json'), 'w') as result:
        result.write(json.dumps(info))
    return info

def hti_phase_trans_analyze(job, jdata=None):

    if_phase_trans = False

    logfile0 = glob.glob(os.path.join(job, '00*', 'hti.out'))[0]
    logfile1 = glob.glob(os.path.join(job, '01*', 'hti.out'))[0]
    logfile2 = glob.glob(os.path.join(job, '02*', 'hti.out'))[0]
    
    log0 = np.loadtxt(logfile0)
    log1 = np.loadtxt(logfile1)
    log2 = np.loadtxt(logfile2)
    print(logfile0, log0)
    print(logfile1, log1)
    print(logfile2, log2)

    msd0 = list(log0[:,-1])
    msd1 = list(log1[:,-1])
    msd2 = list(log2[:,-1])
    
    msd_all = []
    msd_all.extend(msd0)
    msd_all.extend(msd1)
    msd_all.extend(msd2)

    msd_min = min(msd_all)
    msd_max = max(msd_all)

    print('# dpti hti 00 log0')
    print(log0)
    print('# dpti hti 01 log1')
    print(log1)
    print('# dpti hti 02 log2')
    print(log2)
    
    if msd_min < 20 and msd_max > 100:
        if_phase_trans = True

    return if_phase_trans

def _main ():
    parser = argparse.ArgumentParser(
        description="Compute free energy by Hamiltonian TI")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')
    parser_gen.add_argument('-s','--switch', type=str, default = 'one-step',
                            choices = ['one-step', 'two-step', 'three-step', ],
                            help='one-step: switching on DP and switching off spring simultanenously.\
                            two-step: 1 switching on DP, 2 switching off spring.\
                            three-step: 1 switching on soft LJ, 2 switching on DP, 3 switching off spring and soft LJ.')
    parser_gen.add_argument("-z", "--meam", help="whether use meam instead of dp", action="store_true")

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-t','--type', type=str, default = 'helmholtz', 
                             choices=['helmholtz', 'gibbs'], 
                             help='the type of free energy')
    parser_comp.add_argument('-m','--inte-method', type=str, default = 'inte', 
                             choices=['inte', 'mbar'], 
                             help='the method of thermodynamic integration')
    parser_comp.add_argument('-s','--scheme', type=str, default = 'simpson', 
                             help='the numeric integration scheme')
    parser_comp.add_argument('-g', '--pv', type=float, default = None,
                             help='press*vol value override to calculate Gibbs free energy')
    parser_comp.add_argument('-G', '--pv-err', type=float, default = None,
                             help='press*vol error')
    args = parser.parse_args()

    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        output = args.output
        jdata = json.load(open(args.PARAM, 'r'))
        if 'crystal' in jdata and jdata['crystal'] == 'frenkel' :
            print('# gen task with Frenkel\'s Einstein crystal')
        else :
            print('# gen task with Vega\'s Einstein molecule')
        print('output:', output)
        make_tasks(output, jdata, ref='einstein', switch=args.switch, if_meam=args.meam)
    elif args.command == 'compute' :
        compute_task(job=args.JOB, free_energy_type=args.type, method=args.inte_method, scheme=args.scheme, manual_pv=args.pv, manual_pv_err=args.pv_err)
      #   if 'reference' not in jdata :
      #       jdata['reference'] = 'einstein'
    
if __name__ == '__main__' :
    _main()



#!/usr/bin/env python3

import os, sys, json, argparse, glob, shutil
import numpy as np
import scipy.constants as pc

from dpti.lib.utils import create_path
from dpti.lib.utils import copy_file_list
from dpti.lib.utils import block_avg
from dpti.lib.utils import integrate
from dpti.lib.utils import integrate_sys_err
from dpti.lib.utils import parse_seq
from dpti.lib.utils import get_task_file_abspath
from dpti.lib.lammps import get_thermo
from dpti.lib.lammps import get_natoms

from dpti import ti

def _main ():
    parser = argparse.ArgumentParser(
        description="Compute free energy by TI")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command', help = 'valid commands')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-m','--inte-method', type=str, default = 'inte', 
                             choices=['inte', 'mbar'], 
                             help='the method of thermodynamic integration')
    parser_comp.add_argument('-e', '--Eo', type=float, default = 0,
                             help='free energy of starting point')
    parser_comp.add_argument('-E', '--Eo-err', type=float, default = 0,
                             help='The statistical error of the starting free energy')
    parser_comp.add_argument('-t', '--To', type=float, 
                             help='the starting thermodynamic position')
    parser_comp.add_argument('-s', '--scheme', type=str, default = 'simpson',
                             help='the numerical integration scheme')
    parser_comp.add_argument('-S', '--shift', type=float, default = 0.0,
                             help='a constant shift in the energy/mole computation, will be removed from FE')

    parser_comp = subparsers.add_parser('refine', help= 'Refine the grid of a job')
    parser_comp.add_argument('-i', '--input', type=str, required=True,
                             help='input job')
    parser_comp.add_argument('-o', '--output', type=str, required=True,
                             help='output job')
    parser_comp.add_argument('-e', '--error', type=float, required=True,
                             help='the error required')

    args = parser.parse_args()
    return exec_args(args=args, parser=parser)

def exec_args(args, parser=None):
    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        output = args.output
        jdata = json.load(open(args.PARAM, 'r'))
        ti.make_tasks(output, jdata)
    elif args.command == 'compute' :
        job = args.JOB
        jdata = json.load(open(os.path.join(job, 'ti_settings.json'), 'r'))
        equi_conf = get_task_file_abspath(job, jdata['equi_conf'])
        natoms = get_natoms(equi_conf)
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
        nmols = natoms // 3
        if args.inte_method == 'inte' :
            ti.post_tasks(job, jdata, args.Eo, Eo_err = args.Eo_err, To = args.To, natoms = nmols, scheme = args.scheme, shift = args.shift)
        elif args.inte_method == 'mbar' :
            ti.post_tasks_mbar(job, jdata, args.Eo, natoms = nmols)
        else :
            raise RuntimeError('unknow integration method')
    elif args.command == 'refine' :
        ti.refine_task(args.input, args.output, args.error)

if __name__ == '__main__' :
    _main()
        



#!/usr/bin/env python3

import os, sys, json, argparse, glob, shutil
import numpy as np
import scipy.constants as pc

from dpti import einstein
from dpti.lib import lmp
from  dpti.lib import water
import pymbar
from dpti.lib.utils import create_path
from dpti.lib.utils import copy_file_list
from dpti.lib.utils import block_avg
from dpti.lib.utils import integrate_range
# from lib.utils import integrate_sys_err
from dpti.lib.utils import compute_nrefine
from dpti.lib.utils import parse_seq
from dpti.lib.utils import get_task_file_abspath
from dpti.lib.utils import get_first_matched_key_from_dict
from dpti.lib.lammps import get_thermo

def _ff_angle_on(lamb,
                 model, 
                 bparam,
                 sparam) :
    bond_k = bparam['bond_k']
    bond_l = bparam['bond_l']
    angle_k = bparam['angle_k']
    angle_t = bparam['angle_t']
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    sigma_oo = sparam['sigma_oo']
    sigma_oh = sparam['sigma_oh']
    sigma_hh = sparam['sigma_hh']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'pair_style      lj/cut/soft %f %f %f  \n' % (nn, alpha_lj, rcut)
    ret += 'pair_coeff      1 1 ${EPSILON} %f %f\n' % (sigma_oo, activation)
    ret += 'pair_coeff      1 2 ${EPSILON} %f %f\n' % (sigma_oh, activation)
    ret += 'pair_coeff      2 2 ${EPSILON} %f %f\n' % (sigma_hh, activation)
    ret += 'bond_style      harmonic\n'
    ret += 'bond_coeff      1 %f %f\n' % (bond_k, bond_l)
    ret += 'variable        ANGLE_K equal ${LAMBDA}*%.16e\n' % angle_k
    ret += 'angle_style     harmonic\n'
    ret += 'angle_coeff     1 ${ANGLE_K} %f\n' % (angle_t)    
    ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes\n'
    ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON\n'    
    return ret

def _ff_deep_on(lamb,
                 model, 
                 bparam,
                 sparam) :
    bond_k = bparam['bond_k']
    bond_l = bparam['bond_l']
    angle_k = bparam['angle_k']
    angle_t = bparam['angle_t']
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    sigma_oo = sparam['sigma_oo']
    sigma_oh = sparam['sigma_oh']
    sigma_hh = sparam['sigma_hh']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'variable        ONE equal 1\n'
    ret += 'pair_style      hybrid/overlay deepmd %s lj/cut/soft %f %f %f  \n' % (model, nn, alpha_lj, rcut)
    ret += 'pair_coeff      * * deepmd\n'
    ret += 'pair_coeff      1 1 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oo, activation)
    ret += 'pair_coeff      1 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oh, activation)
    ret += 'pair_coeff      2 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_hh, activation)
    ret += 'bond_style      harmonic\n'
    ret += 'bond_coeff      1 %f %f\n' % (bond_k, bond_l)
    ret += 'angle_style     harmonic\n'
    ret += 'angle_coeff     1 %f %f\n' % (angle_k, angle_t)    
    ret += 'fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA\n'
    ret += 'compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE\n'
    return ret

def _ff_bond_angle_off(lamb,
                       model, 
                       bparam,
                       sparam) :
    bond_k = bparam['bond_k']
    bond_l = bparam['bond_l']
    angle_k = bparam['angle_k']
    angle_t = bparam['angle_t']
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    sigma_oo = sparam['sigma_oo']
    sigma_oh = sparam['sigma_oh']
    sigma_hh = sparam['sigma_hh']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        INV_LAMBDA equal 1-${LAMBDA}\n'
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'variable        INV_EPSILON equal -${EPSILON}\n'
    ret += 'pair_style      hybrid/overlay deepmd %s lj/cut/soft %f %f %f  \n' % (model, nn, alpha_lj, rcut)
    ret += 'pair_coeff      * * deepmd\n'
    ret += 'pair_coeff      1 1 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oo, activation)
    ret += 'pair_coeff      1 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_oh, activation)
    ret += 'pair_coeff      2 2 lj/cut/soft ${EPSILON} %f %f\n' % (sigma_hh, activation)
    ret += 'variable        BOND_K equal %.16e\n' % (bond_k * (1-lamb))
    ret += 'bond_style      harmonic\n'
    ret += 'bond_coeff      1 ${BOND_K} %f\n' % (bond_l)
    ret += 'variable        ANGLE_K equal %.16e\n' % (angle_k * (1-lamb))
    ret += 'angle_style     harmonic\n'
    ret += 'angle_coeff     1 ${ANGLE_K} %f\n' % (angle_t)    
    ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes\n'
    ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON\n'    
    return ret


def _gen_lammps_input (step,
                       conf_file, 
                       mass_map,
                       lamb,
                       model,
                       bparam,
                       sparam,
                       nsteps,
                       dt,
                       ens,
                       temp,
                       pres = 1.0, 
                       tau_t = 0.1,
                       tau_p = 0.5,
                       prt_freq = 100, 
                       copies = None) :
    ret = ''
    ret += 'clear\n'
    ret += '# --------------------- VARIABLES-------------------------\n'
    ret += 'variable        NSTEPS          equal %d\n' % nsteps
    ret += 'variable        THERMO_FREQ     equal %d\n' % prt_freq
    ret += 'variable        DUMP_FREQ       equal %d\n' % prt_freq
    ret += 'variable        TEMP            equal %f\n' % temp
    ret += 'variable        PRES            equal %f\n' % pres
    ret += 'variable        TAU_T           equal %f\n' % tau_t
    ret += 'variable        TAU_P           equal %f\n' % tau_p
    ret += 'variable        LAMBDA          equal %.10e\n' % lamb
    ret += '# ---------------------- INITIALIZAITION ------------------\n'
    ret += 'units           metal\n'
    ret += 'boundary        p p p\n'
    ret += 'atom_style      angle\n'
    ret += '# --------------------- ATOM DEFINITION ------------------\n'
    ret += 'box             tilt large\n'
    ret += 'read_data       %s\n' % conf_file
    if copies is not None :
        ret += 'replicate       %d %d %d\n' % (copies[0], copies[1], copies[2])
    ret += 'change_box      all triclinic\n'
    for jj in range(len(mass_map)) :
        ret+= "mass            %d %f\n" %(jj+1, mass_map[jj])
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    if step == 'angle_on' :
        ret += _ff_angle_on(lamb, model, bparam, sparam)
    elif step == 'deep_on':
        ret += _ff_deep_on(lamb, model, bparam, sparam)
    elif step == 'bond_angle_off':
        ret += _ff_bond_angle_off(lamb, model, bparam, sparam)
    ret += 'special_bonds   lj/coul 1 1 1 angle no\n'
    ret += '# --------------------- MD SETTINGS ----------------------\n'    
    ret += 'neighbor        1.0 bin\n'
    ret += 'timestep        %s\n' % dt
    ret += 'thermo          ${THERMO_FREQ}\n'
    ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol ebond eangle c_e_diff[1]\n'
    ret += 'thermo_modify   format 9 %.16e\n'
    ret += 'thermo_modify   format 10 %.16e\n'
    ret += 'thermo_modify   format 11 %.16e\n'
    ret += '# dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z\n'
    if ens == 'nvt' :
        ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
    elif ens == 'npt-iso' or ens == 'npt':
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'nve' :
        ret += 'fix             1 all nve\n'
    else :
        raise RuntimeError('unknow ensemble %s\n' % ens)        
    ret += 'fix             mzero all momentum 10 linear 1 1 1\n'
    ret += '# --------------------- INITIALIZE -----------------------\n'    
    ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
    ret += 'velocity        all zero linear\n'
    ret += '# --------------------- RUN ------------------------------\n'    
    ret += 'run             ${NSTEPS}\n'    
    ret += 'write_data      out.lmp\n'

    return ret


def _make_tasks(iter_name, jdata, step) :
    if step == 'angle_on' :
        all_lambda = parse_seq(jdata['lambda_angle_on'])
        protect_eps = jdata['protect_eps']    
        if all_lambda[0] == 0 :
            all_lambda[0] += protect_eps
    elif step == 'deep_on' :
        all_lambda = parse_seq(jdata['lambda_deep_on'])
        protect_eps = jdata['protect_eps']    
        if all_lambda[0] == 0 :
            all_lambda[0] += protect_eps
    elif step == 'bond_angle_off' :
        all_lambda = parse_seq(jdata['lambda_bond_angle_off'])
        protect_eps = jdata['protect_eps']    
        if all_lambda[-1] == 1 :
            all_lambda[-1] -= protect_eps
    equi_conf = jdata['equi_conf']
    equi_conf = os.path.abspath(equi_conf)
    model = jdata['model']
    model = os.path.abspath(model)
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    nsteps = jdata['nsteps']
    # dt = jdata['dt']
    # timestep = jdata['timestep']
    timestep = get_first_matched_key_from_dict(jdata, ['timestep', 'dt'])
    bparam = jdata['bond_param']
    sparam = jdata['soft_param']
    # stat_freq = jdata['stat_freq']
    # thermo_freq = jdata['thermo_freq']
    thermo_freq = get_first_matched_key_from_dict(jdata, ['thermo_freq', 'stat_freq'])
    ens = jdata['ens']
    temp = jdata['temp']
    pres = jdata['pres']
    tau_t = jdata['tau_t']
    tau_p = jdata['tau_p']
    copies = None
    if 'copies' in jdata :
        copies = jdata['copies']

    create_path(iter_name)
    cwd = os.getcwd()
    os.chdir(iter_name)
    os.symlink(os.path.join('..', 'in.json'), 'in.json')
    os.symlink(os.path.join('..', 'conf.lmp'), 'orig.lmp')
    os.symlink(os.path.join('..', 'graph.pb'), 'graph.pb')
    with open('orig.lmp', 'r') as f:
        lines = water.add_bonds(f.read().split('\n'))
    with open('conf.lmp', 'w') as c:
        c.write('\n'.join(lines))
    os.chdir(cwd)
    for idx in range(len(all_lambda)) :
        work_path = os.path.join(iter_name, 'task.%06d' % idx)
        create_path(work_path)
        os.chdir(work_path)
        os.symlink(os.path.join('..', 'conf.lmp'), 'conf.lmp')
        os.symlink(os.path.join('..', 'graph.pb'), 'graph.pb')
        lmp_str \
            = _gen_lammps_input(step,
                                'conf.lmp', 
                                mass_map, 
                                all_lambda[idx],
                                'graph.pb',
                                bparam,
                                sparam,
                                nsteps, 
                                timestep, 
                                ens, 
                                temp, 
                                pres, 
                                tau_t = tau_t,
                                tau_p = tau_p,
                                prt_freq = thermo_freq, 
                                copies = copies)
        with open('in.lammps', 'w') as fp :
            fp.write(lmp_str)
        with open('lambda.out', 'w') as fp :
            fp.write(str(all_lambda[idx]))
        os.chdir(cwd)

def _refine_tasks(from_task, to_task, err, step) :
    from_task = os.path.abspath(from_task)
    to_task = os.path.abspath(to_task)

    from_ti = os.path.join(from_task, 'hti.out')
    if not os.path.isfile(from_ti) :
        raise RuntimeError("cannot find file %s, task should be computed befor refined" % from_ti)
    tmp_array = np.loadtxt(from_ti)
    all_t = tmp_array[:,0]
    integrand = tmp_array[:,1]
    ntask = all_t.size
    
    interval_nrefine = compute_nrefine(all_t, integrand, err)

    refined_t = []
    back_map = []
    for ii in range(0, ntask-1) :
        refined_t.append(all_t[ii])
        back_map.append(ii)
        hh = (all_t[ii+1] - all_t[ii]) / interval_nrefine[ii]
        for jj in range(1, interval_nrefine[ii]) :
            refined_t.append(all_t[ii] + jj * hh)
            back_map.append(-1)
    refined_t.append(all_t[-1])
    back_map.append(ntask-1)
    
    from_json = os.path.join(from_task, 'in.json')
    to_json = os.path.join(to_task, 'in.json')
    from_jdata = json.load(open(from_json))
    to_jdata = from_jdata

    to_jdata['orig_task'] = from_task
    to_jdata['back_map'] = back_map
    to_jdata['refine_error'] = err
    if step == 'angle_on' :
        to_jdata['lambda_angle_on'] = refined_t
        to_jdata['lambda_angle_on_back_map'] = back_map
    elif step == 'deep_on' :
        to_jdata['lambda_deep_on'] = refined_t
        to_jdata['lambda_deep_on_back_map'] = back_map
    elif step == 'bond_angle_off' :
        to_jdata['lambda_bond_angle_off'] = refined_t
        to_jdata['lambda_bond_angle_off_back_map'] = back_map
    else :
        raise RuntimeError('unknow step')

    _make_tasks(to_task, to_jdata, step)

    from_task_list = glob.glob(os.path.join(from_task, 'task.[0-9]*'))
    from_task_list.sort()
    to_task_list = glob.glob(os.path.join(to_task, 'task.[0-9]*'))
    to_task_list.sort()
    assert(len(from_task_list) == ntask)
    assert(len(to_task_list) == len(refined_t))

    for ii in range(len(to_task_list)) :
        if back_map[ii] < 0 : 
            continue
        for jj in ['data', 'log.lammps'] :
            shutil.copyfile(
                os.path.join(from_task_list[back_map[ii]], jj), 
                os.path.join(to_task_list[ii], jj), 
            )
        with open(os.path.join(to_task_list[ii], 'from.dir'), 'w') as fp:
            fp.write(from_task_list[back_map[ii]])    


def make_tasks(iter_name, jdata) :
    equi_conf = os.path.abspath(jdata['equi_conf'])
    model = os.path.abspath(jdata['model'])

    create_path(iter_name)
    copied_conf = os.path.join(os.path.abspath(iter_name), 'conf.lmp')
    shutil.copyfile(equi_conf, copied_conf)
    jdata['equi_conf'] = 'conf.lmp'
    linked_model = os.path.join(os.path.abspath(iter_name), 'graph.pb')
    shutil.copyfile(model, linked_model)
    jdata['model'] = 'graph.pb'

    cwd = os.getcwd()
    os.chdir(iter_name)    
    with open('in.json', 'w') as fp:
        json.dump(jdata, fp, indent=4)
    os.chdir(cwd)
    subtask_name = os.path.join(iter_name, '00.angle_on')
    _make_tasks(subtask_name, jdata, 'angle_on')
    subtask_name = os.path.join(iter_name, '01.deep_on')
    _make_tasks(subtask_name, jdata, 'deep_on')
    subtask_name = os.path.join(iter_name, '02.bond_angle_off')
    _make_tasks(subtask_name, jdata, 'bond_angle_off')


def refine_tasks(from_task, to_task, err) :
    jdata = json.load(open(os.path.join(from_task, 'in.json')))    
    equi_conf = get_task_file_abspath(from_task, jdata['equi_conf'])
    model = get_task_file_abspath(from_task, jdata['model'])

    create_path(to_task)
    copied_conf = os.path.join(os.path.abspath(to_task), 'conf.lmp')
    shutil.copyfile(equi_conf, copied_conf)
    jdata['equi_conf'] = 'conf.lmp'
    linked_model = os.path.join(os.path.abspath(to_task), 'graph.pb')
    shutil.copyfile(model, linked_model)
    jdata['model'] = 'graph.pb'
    jdata['orig_task'] = from_task
    jdata['refine_error'] = err

    cwd = os.getcwd()
    os.chdir(to_task)
    with open('in.json', 'w') as fp:
        json.dump(jdata, fp, indent=4)
    os.chdir(cwd)

    from_name = os.path.join(from_task, '00.angle_on')
    to_name = os.path.join(to_task, '00.angle_on')
    _refine_tasks(from_name, to_name, err, 'angle_on')
    from_name = os.path.join(from_task, '01.deep_on')
    to_name = os.path.join(to_task, '01.deep_on')
    _refine_tasks(from_name, to_name, err, 'deep_on')
    from_name = os.path.join(from_task, '02.bond_angle_off')
    to_name = os.path.join(to_task, '02.bond_angle_off')
    _refine_tasks(from_name, to_name, err, 'bond_angle_off')


def _compute_thermo(fname, natoms, stat_skip, stat_bsize) :
    data = get_thermo(fname)
    ea, ee = block_avg(data[:, 3], skip = stat_skip, block_size = stat_bsize)
    ha, he = block_avg(data[:, 4], skip = stat_skip, block_size = stat_bsize)
    ta, te = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
    pa, pe = block_avg(data[:, 6], skip = stat_skip, block_size = stat_bsize)
    va, ve = block_avg(data[:, 7], skip = stat_skip, block_size = stat_bsize)
    thermo_info = {}
    thermo_info['p'] = pa
    thermo_info['p_err'] = pe
    thermo_info['v'] = va / natoms
    thermo_info['v_err'] = ve / np.sqrt(natoms)
    thermo_info['e'] = ea / natoms
    thermo_info['e_err'] = ee / np.sqrt(natoms)
    thermo_info['h'] = ha / natoms
    thermo_info['h_err'] = he / np.sqrt(natoms)
    thermo_info['t'] = ta
    thermo_info['t_err'] = te
    unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
    thermo_info['pv'] = pa * va * unit_cvt / natoms
    thermo_info['pv_err'] = pe * va * unit_cvt  / np.sqrt(natoms)
    return thermo_info

def _post_tasks(iter_name, step, natoms, scheme = 's') :
    jdata = json.load(open(os.path.join(iter_name, 'in.json')))
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)
    
    all_lambda = []
    all_bd_a = []
    all_bd_e = []
    all_ag_a = []
    all_ag_e = []
    all_dp_a = []
    all_dp_e = []

    for ii in all_tasks :
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        bd_a, bd_e = block_avg(data[:, 8], skip = stat_skip, block_size = stat_bsize)
        ag_a, ag_e = block_avg(data[:, 9], skip = stat_skip, block_size = stat_bsize)
        dp_a, dp_e = block_avg(data[:,10], skip = stat_skip, block_size = stat_bsize)
        bd_a /= natoms
        ag_a /= natoms
        dp_a /= natoms
        bd_e /= np.sqrt(natoms)
        ag_e /= np.sqrt(natoms)
        dp_e /= np.sqrt(natoms)
        lmda_name = os.path.join(ii, 'lambda.out')
        ll = float(open(lmda_name).read())
        all_lambda.append(ll)
        all_bd_a.append(bd_a)
        all_bd_e.append(bd_e)
        all_ag_a.append(ag_a)
        all_ag_e.append(ag_e)
        all_dp_a.append(dp_a)
        all_dp_e.append(dp_e)

    all_lambda = np.array(all_lambda)
    all_bd_a = np.array(all_bd_a)
    all_bd_e = np.array(all_bd_e)
    all_ag_a = np.array(all_ag_a)
    all_ag_e = np.array(all_ag_e)
    all_dp_a = np.array(all_dp_a)
    all_dp_e = np.array(all_dp_e)
    if step == 'angle_on' :        
        de = all_ag_a / all_lambda + all_dp_a
        all_err = np.sqrt(np.square(all_ag_e / all_lambda) +
                          np.square(all_dp_e))
    elif step == 'deep_on' :
        de = all_dp_a
        all_err = all_dp_e
    elif step == 'bond_angle_off' :
        de = - (all_bd_a + all_ag_a) / (1 - all_lambda) + all_dp_a
        all_err = np.sqrt(np.square(all_bd_e / (1 - all_lambda)) + 
                          np.square(all_ag_e / (1 - all_lambda)) + 
                          np.square(all_dp_e))

    all_print = []
    # all_print.append(np.arange(len(all_lambda)))
    all_print.append(all_lambda)
    all_print.append(de)
    all_print.append(all_err)
    all_print = np.array(all_print)
    np.savetxt(os.path.join(iter_name, 'hti.out'), 
               all_print.T, 
               fmt = '%.8e', 
               header = 'lmbda dU dU_err')

    new_lambda, i, i_e, s_e = integrate_range(all_lambda, de, all_err, scheme = scheme)
    if new_lambda[-1] != all_lambda[-1] :
        if new_lambda[-1] == all_lambda[-2]:
            _, i1, i_e1, s_e1 = integrate_range(all_lambda[-2:], de[-2:], all_err[-2:], scheme='t')
            diff_e = i[-1] + i1[-1]
            err = np.linalg.norm([s_e[-1], s_e1[-1]])
            sys_err = i_e[-1] + i_e1[-1]
        else :
            raise RuntimeError("lambda does not match!")
    else:
        diff_e = i[-1]
        err = s_e[-1]
        sys_err = i_e[-1]

    # diff_e, err = integrate(all_lambda, de, all_err)
    # sys_err = integrate_sys_err(all_lambda, de)

    thermo_info = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'), 
                                  natoms,
                                  stat_skip, stat_bsize)

    return diff_e, [err, sys_err], thermo_info


def _post_tasks_mbar(iter_name, step, natoms) :
    jdata = json.load(open(os.path.join(iter_name, 'in.json')))
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    temp = jdata['temp']
    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)

    all_lambda = []
    for ii in all_tasks :
        lmda_name = os.path.join(ii, 'lambda.out')
        ll = float(open(lmda_name).read())
        all_lambda.append(ll)
    all_lambda = np.array(all_lambda)
    nlambda = all_lambda.size

    ukn = np.array([])
    nk = []
    kt_in_ev = pc.Boltzmann * temp / pc.electron_volt
    for idx,ii in enumerate(all_tasks) :
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        bd_e = data[stat_skip:, 8]/kt_in_ev
        ag_e = data[stat_skip:, 9]/kt_in_ev
        dp_e = data[stat_skip:,10]/kt_in_ev
        if step == 'angle_on' :        
            de = ag_e / all_lambda[idx] + dp_e
        elif step == 'deep_on' :
            de = dp_e
        elif step == 'bond_angle_off' :
            de = -(bd_e + ag_e) / (1 - all_lambda[idx]) + dp_e
        else :
            raise RuntimeError("unknow step")
        nk.append(de.size)
        block_u = []
        for ll in all_lambda :
            if step == 'angle_on' or 'deep_on':
                block_u.append(de * ll)
            else :
                block_u.append(-de * (1-ll))
        block_u = np.reshape(block_u, [nlambda, -1])
        if ukn.size == 0 :
            ukn = block_u 
        else :
            ukn = np.concatenate((ukn, block_u), axis = 1)

    nk = np.array(nk)
    mbar = pymbar.MBAR(ukn, nk)
    Deltaf_ij, dDeltaf_ij, Theta_ij = mbar.getFreeEnergyDifferences()
    Deltaf_ij = Deltaf_ij / natoms
    dDeltaf_ij = dDeltaf_ij / np.sqrt(natoms)

    diff_e = Deltaf_ij[0,-1] * kt_in_ev
    err = dDeltaf_ij[0,-1] * kt_in_ev
    sys_err = 0

    thermo_info = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'), 
                                  natoms,
                                  stat_skip, stat_bsize)

    return diff_e, [err, sys_err], thermo_info


def _print_thermo_info(info) :
    ptr = '# thermodynamics (normalized by nmols)\n'
    ptr += '# E (err)  [eV]:  %20.8f %20.8f\n' % (info['e'], info['e_err'])
    ptr += '# H (err)  [eV]:  %20.8f %20.8f\n' % (info['h'], info['h_err'])
    ptr += '# T (err)   [K]:  %20.8f %20.8f\n' % (info['t'], info['t_err'])
    ptr += '# P (err) [bar]:  %20.8f %20.8f\n' % (info['p'], info['p_err'])
    ptr += '# V (err) [A^3]:  %20.8f %20.8f\n' % (info['v'], info['v_err'])
    ptr += '# PV(err)  [eV]:  %20.8f %20.8f' % (info['pv'], info['pv_err'])
    print(ptr)

def spring_inte(temp, kk, r0) :
    kto2k = pc.Boltzmann * temp / (2. * kk * pc.electron_volt / (pc.angstrom * pc.angstrom))
    # print((r0 * pc.angstrom), np.sqrt(kto2k))
    return 4 * np.pi * np.sqrt(2 * np.pi * kto2k) * ((r0 * pc.angstrom) ** 2 + kto2k)

def compute_ideal_mol(iter_name) :
    jdata = json.load(open(os.path.join(iter_name, 'in.json')))
    ens = jdata['ens']
    mass_map = jdata['mass_map']
    conf_lines = open(os.path.join(iter_name, 'orig.lmp')).read().split('\n')
    data_sys = lmp.system_data(conf_lines)
    vol = np.linalg.det(data_sys['cell'])
    temp = jdata['temp']    
    kk = jdata['bond_param']['bond_k']
    ll = jdata['bond_param']['bond_l']
    if 'copies' in jdata :
        ncopies = np.prod(jdata['copies'])
    else :
        ncopies = 1
    natom_vec = [ii * ncopies for ii in data_sys['atom_numbs']]

    # kinetic contribution
    fe = 0
    for ii in range(len(natom_vec)) :
        mass = mass_map[ii]
        natoms = natom_vec[ii]
        lambda_k = einstein.compute_lambda(temp, mass)
        fe += 3 * natoms * np.log(lambda_k)
    natoms_o = natom_vec[0]
    natoms_h = 2 * natoms_o
    natoms = natoms_o + natoms_h
    assert(natoms == sum(natom_vec))
    # spring contribution
    lambda_s = einstein.compute_spring(temp, kk * 2.0)
    lambda_s1 = spring_inte(temp, kk, ll)
    fe -= natoms_o * np.log((vol * (pc.angstrom**3)))
    # print((1/lambda_s))
    # fe += 3 * natoms_h * np.log(lambda_s)
    fe -= natoms_h * np.log(lambda_s1)
    # N!
    fe += natoms_o * np.log(natoms_o) - natoms_o + 0.5 * np.log(2. * np.pi * natoms_o) 
    fe += natoms_h * np.log(np.sqrt(2))
    # plus PV
    if 'npt' in ens :
        fe += natoms_o + 5./6. * natoms_h
    # to kbT log Z
    fe *= pc.Boltzmann * temp / pc.electron_volt
    return fe / natoms_o

def post_tasks(iter_name, natoms, method = 'inte', scheme = 's') :
    subtask_name = os.path.join(iter_name, '00.angle_on')
    fe = compute_ideal_mol(subtask_name)
    print('# fe of ideal mol: %20.12f' % fe)
#    print('# fe of ideal gas: %20.12f' % (einstein.ideal_gas_fe(subtask_name) * 3))
    if method == 'inte' :
        e0, err0, tinfo0 = _post_tasks(subtask_name, 'angle_on', natoms, scheme=scheme)
    elif method == 'mbar' :
        e0, err0, tinfo0 = _post_tasks_mbar(subtask_name, 'angle_on', natoms)
    print('# fe of angle_on : %20.12f  %10.3e %10.3e' % (e0, err0[0], err0[1]))
    # _print_thermo_info(tinfo)
    # print(e, err)
    subtask_name = os.path.join(iter_name, '01.deep_on')
    if method == 'inte' :
        e1, err1, tinfo1 = _post_tasks(subtask_name, 'deep_on', natoms, scheme=scheme)
    elif method == 'mbar' :
        e1, err1, tinfo1 = _post_tasks_mbar(subtask_name, 'deep_on', natoms)
    print('# fe of deep_on  : %20.12f  %10.3e %10.3e' % (e1, err1[0], err1[1]))
    # _print_thermo_info(tinfo)
    # print(e, err)
    subtask_name = os.path.join(iter_name, '02.bond_angle_off')
    if method == 'inte' :
        e2, err2, tinfo2 = _post_tasks(subtask_name, 'bond_angle_off', natoms, scheme=scheme)
    elif method == 'mbar' :
        e2, err2, tinfo2 = _post_tasks_mbar(subtask_name, 'bond_angle_off', natoms)
    print('# fe of bond_off : %20.12f  %10.3e %10.3e' % (e2, err2[0], err2[1]))
    # _print_thermo_info(tinfo)
    # print(e, err)
    fe = fe + e0 + e1 + e2
    err = np.sqrt(np.square(err0[0]) + np.square(err1[0]) + np.square(err2[0]))
    sys_err = ((err0[1]) + (err1[1]) + (err2[1]))
    return fe, [err,sys_err], tinfo2


def _main ():
    parser = argparse.ArgumentParser(
        description="Compute free energy by Hamiltonian TI")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-t','--type', type=str, default = 'helmholtz', 
                             choices=['helmholtz', 'gibbs'], 
                             help='the type of free energy')
    parser_comp.add_argument('-m','--inte-method', type=str, default = 'inte', 
                             choices=['inte', 'mbar'], 
                             help='the method of thermodynamic integration')
    parser_comp.add_argument('-s','--scheme', type=str, default = 'simpson', 
                             help='the numeric integration scheme')
    parser_comp.add_argument('-g', '--pv', type=float, default = None,
                             help='press*vol value override to calculate Gibbs free energy')
    parser_comp.add_argument('-G', '--pv-err', type=float, default = None,
                             help='press*vol error')

    parser_comp = subparsers.add_parser('refine', help= 'Refine the grid of a job')
    parser_comp.add_argument('-i', '--input', type=str, required=True,
                             help='input job')
    parser_comp.add_argument('-o', '--output', type=str, required=True,
                             help='output job')
    parser_comp.add_argument('-e', '--error', type=float, required=True,
                             help='the error required')

    args = parser.parse_args()
    return exec_args(args=args, parser=None)

def exec_args(args, parser):
    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        output = args.output
        with open(args.PARAM, 'r') as j:
            jdata = json.load(j)
        make_tasks(output, jdata)
    if args.command == 'refine' :
        refine_tasks(args.input, args.output, args.error)
    elif args.command == 'compute' :
        with open(os.path.join(args.JOB, 'conf.lmp'), 'r') as conf_lmp:
            # fp_conf = open(os.path.join(args.JOB, 'conf.lmp'))
            sys_data = lmp.to_system_data(conf_lmp.read().split('\n'))
        natoms = sum(sys_data['atom_numbs'])
        with open(os.path.join(args.JOB, 'in.json'), 'r') as j:
            jdata = json.load(j)

        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
        nmols = natoms // 3
        fe, fe_err, thermo_info = post_tasks(args.JOB, nmols, method = args.inte_method)
        _print_thermo_info(thermo_info)
        print ('# numb atoms: %d' % natoms)
        print ('# numb  mols: %d' % nmols)        
        print_format = '%20.12f  %10.3e  %10.3e'
        # if args.type == 'helmholtz' :
        print('# Helmholtz free ener per mol (err) [eV]:')
        print(print_format % (fe, fe_err[0], fe_err[1]))
        if args.type == 'gibbs':
            if args.pv is not None:
                pv = args.pv
                print(f"# use manual pv=={pv}")
            else:
                pv = thermo_info['pv']
            if args.pv_err is not None:
                pv_err = args.pv_err
                print(f"# use manual pv_err=={pv_err}")
            else:
                pv_err = thermo_info['pv_err']
            e1 = fe + pv
            e1_err = np.sqrt(fe_err[0]**2 + pv_err**2)
            print('# Gibbs free ener per mol (err) [eV]:')
            print(print_format % (e1, e1_err, fe_err[1]))
    
if __name__ == '__main__' :
    _main()



#%%
import os, textwrap, json, shutil
import numpy as np
import unittest
# from potential_common import soft_param, soft_param_three_element, meam_model
# from dpti.lib.lammps import get_natoms, get_thermo, get_last_dump
# from dpti.lib.dump import from_system_data
from dpti.hti import _gen_lammps_input
from numpy.testing import assert_almost_equal
from unittest.mock import MagicMock, patch, PropertyMock

from dpti import hti_ice
from dpti.lib.utils import get_file_md5

class TestHtiIceGenLammpsInput(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None
        # with open('benchmark_hti_ice/hti_ice.json', 'r') as f:
        #     self.jdata = json.load(f)

    @patch('numpy.random')
    def test_hti_ice_tasks(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        args = MagicMock(output='tmp_hti_ice/new_job/', 
            switch='three-step', 
            command='gen',
            PARAM='benchmark_hti_ice/hti_ice.json'
        )
        hti_ice.exec_args(args=args, parser=None)
        check_file_list = [ 
            'conf.lmp', 
            '00.lj_on/task.000003/conf.lmp',
            '00.lj_on/task.000003/in.lammps',
            '01.deep_on/task.000004/in.lammps',
            '02.spring_off/task.000005/in.lammps'
        ]
        for file in check_file_list:
            f1 = os.path.join('benchmark_hti_ice/new_job/', file)
            f2 = os.path.join('tmp_hti_ice/new_job/', file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_hti_ice_old_json_tasks(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        args = MagicMock(output='tmp_hti_ice/old_json_job/', 
            switch='three-step', 
            command='gen',
            PARAM='benchmark_hti_ice/hti_ice.json.old'
        )
        hti_ice.exec_args(args=args, parser=None)
        check_file_list = [ 
            'conf.lmp', 
            '00.lj_on/task.000003/conf.lmp',
            '00.lj_on/task.000003/in.lammps',
            '01.deep_on/task.000004/in.lammps',
            '02.spring_off/task.000005/in.lammps'
        ]
        for file in check_file_list:
            f1 = os.path.join('benchmark_hti_ice/new_job/', file)
            f2 = os.path.join('tmp_hti_ice/old_json_job/', file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_hti_ice/')



#!/usr/bin/python3 

import warnings
import numpy as np

def regulate_poscar(poscar_in, poscar_out) :
    with open(poscar_in, 'r') as fp:
        lines = fp.read().split('\n')
    names = lines[5].split()
    counts = [int(ii) for ii in lines[6].split()]
    assert(len(names) == len(counts))
    uniq_name = []
    for ii in names :
        if not (ii in uniq_name) :
            uniq_name.append(ii)
    uniq_count = np.zeros(len(uniq_name), dtype = int)
    for nn,cc in zip(names,counts) :
        uniq_count[uniq_name.index(nn)] += cc
    natoms = np.sum(uniq_count)
    posis = lines[8:8+natoms]
    all_lines = []
    for ele in uniq_name:
        ele_lines = []
        for ii in posis :
            ele_name = ii.split()[-1]
            if ele_name == ele :
                ele_lines.append(ii) 
        all_lines += ele_lines
    all_lines.append('')
    ret = lines[0:5]
    ret.append(" ".join(uniq_name))
    ret.append(" ".join([str(ii) for ii in uniq_count]))
    ret.append("Direct")
    ret += all_lines
    with open(poscar_out, 'w') as fp:
        fp.write("\n".join(ret))

def sort_poscar(poscar_in, poscar_out, new_names) :
    with open(poscar_in, 'r') as fp:
        lines = fp.read().split('\n')
    names = lines[5].split()
    counts = [int(ii) for ii in lines[6].split()]
    new_counts = np.zeros(len(counts), dtype = int)
    for nn,cc in zip(names,counts) :
        new_counts[new_names.index(nn)] += cc
    natoms = np.sum(new_counts)
    posis = lines[8:8+natoms]
    all_lines = []
    for ele in new_names:
        ele_lines = []
        for ii in posis :
            ele_name = ii.split()[-1]
            if ele_name == ele :
                ele_lines.append(ii) 
        all_lines += ele_lines
    all_lines.append('')
    ret = lines[0:5]
    ret.append(" ".join(new_names))
    ret.append(" ".join([str(ii) for ii in new_counts]))
    ret.append("Direct")
    ret += all_lines
    with open(poscar_out, 'w') as fp:
        fp.write("\n".join(ret))    

def perturb_xz (poscar_in, poscar_out, pert = 0.01) :
    with open(poscar_in, 'r') as fp:
        lines = fp.read().split('\n')
    zz = lines[4]
    az = [float(ii) for ii in zz.split()]
    az[0] += pert
    zz = [str(ii) for ii in az]
    zz = " ".join(zz)
    lines[4] = zz
    with open(poscar_out, 'w') as fp:
        fp.write("\n".join(lines))

def reciprocal_box(box) :
    rbox = np.linalg.inv(box)
    rbox = rbox.T
    # rbox = rbox / np.linalg.det(box)
    # print(np.matmul(box, rbox.T))
    # print(rbox)
    return rbox

def _poscar_natoms(lines) :
    numb_atoms = 0
    for ii in lines[6].split() :
        numb_atoms += int(ii)
    return numb_atoms

def _poscar_scale_direct (str_in, scale) :
    lines = str_in.copy()
    numb_atoms = _poscar_natoms(lines)
    pscale = float(lines[1])
    pscale = pscale * scale
    lines[1] = str(pscale) + "\n"
    return lines

def _poscar_scale_cartesian (str_in, scale) :
    lines = str_in.copy()
    numb_atoms = _poscar_natoms(lines)
    # scale box
    for ii in range(2,5) :
        boxl = lines[ii].split()
        boxv = [float(ii) for ii in boxl]
        boxv = np.array(boxv) * scale
        lines[ii] = "%.16e %.16e %.16e\n" % (boxv[0], boxv[1], boxv[2])
    # scale coord
    for ii in range(8, 8+numb_atoms) :
        cl = lines[ii].split()
        cv = [float(ii) for ii in cl]
        cv = np.array(cv) * scale
        lines[ii] = "%.16e %.16e %.16e\n" % (cv[0], cv[1], cv[2])
    return lines    

def poscar_natoms(poscar_in) :
    with open(poscar_in, 'r') as fin :
        lines = list(fin)
    return _poscar_natoms(lines)

def poscar_scale (poscar_in, poscar_out, scale) :
    with open(poscar_in, 'r') as fin :
        lines = list(fin)
    if 'D' == lines[7][0] or 'd' == lines[7][0] : 
        lines = _poscar_scale_direct(lines, scale)
    elif 'C' == lines[7][0] or 'c' == lines[7][0] : 
        lines = _poscar_scale_cartesian(lines, scale)
    else :
        raise RuntimeError("Unknow poscar coord style at line 7: %s" % lines[7])
    with open(poscar_out, 'w') as fout:
        fout.write("".join(lines))

def poscar_vol (poscar_in) :
    with open(poscar_in, 'r') as fin :
        lines = list(fin)
    box = []
    for ii in range(2,5) :
        words = lines[ii].split()
        vec = [float(jj) for jj in words]
        box.append(vec)
    scale = float(lines[1].split()[0])
    box = np.array(box)
    box *= scale
    return np.linalg.det(box)




#!/usr/bin/env python3

import os, sys, json, argparse, glob
import numpy as np
import scipy.constants as pc
# from . import lib
from dpti.lib import lmp
from dpti.lib.utils import get_first_matched_key_from_dict
# from lib import lmp

def compute_lambda(temp, mass) :
    ret = 2. * np.pi * mass * (1e-3 / pc.Avogadro) * pc.Boltzmann * temp / (pc.Planck * pc.Planck)
    # print('!!', mass, 1./np.sqrt(ret))
    return 1./np.sqrt(ret)    

def compute_spring(temp, spring_k) :
    ret = (0.5 * spring_k * pc.electron_volt / (pc.angstrom * pc.angstrom)) / (pc.Boltzmann * temp * np.pi) 
    return np.sqrt(ret)

def ideal_gas_fe(job):
    with open(os.path.join(job, 'in.json'), 'r') as f:
        jdata = json.load(f)
    equi_conf = jdata['equi_conf']
    cwd = os.getcwd()
    os.chdir(job)
    assert(os.path.isfile(equi_conf)),equi_conf
    equi_conf = os.path.abspath(equi_conf)
    os.chdir(cwd)
    temp = jdata['temp']
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    if 'copies' in jdata :
        ncopies = np.prod(jdata['copies'])
    else :
        ncopies = 1

    with open(equi_conf, 'r') as f:
        sys_data = lmp.to_system_data(f.read().split('\n'))
    vol = np.linalg.det(sys_data['cell'])
    natoms = [ii * ncopies for ii in sys_data['atom_numbs']]
    Lambda_k = [compute_lambda(temp, ii) for ii in mass_map]    
    fe = 0
    for idx,ii in enumerate(natoms) :
        # kinetic contrib
        # print('```', idx, ii)
        if ii > 0:
            rho = ii / (vol * (pc.angstrom**3))
            fe += ii * np.log(rho * (Lambda_k[idx] ** 3)) 
            fe -= ii
            fe += 0.5 * np.log(2. * np.pi * ii)
    fe *= pc.Boltzmann * temp / pc.electron_volt
    fe /= np.sum(natoms)
    return fe

def free_energy(job) :
    # vega style free energy
    # print(898, job, os.getcwd())
    with open(os.path.join(job, 'in.json')) as f:
        jdata = json.load(f)
    # jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))    
    equi_conf = jdata['equi_conf']
    cwd = os.getcwd()
    os.chdir(job)
    assert(os.path.isfile(equi_conf))
    equi_conf = os.path.abspath(equi_conf)
    os.chdir(cwd)
    temp = jdata['temp']
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    spring_k = jdata['spring_k']
    if type(spring_k) is not list:
        m_spring_k = []
        for ii in mass_map :
            m_spring_k.append(spring_k * ii)
        # spring_k = spring_k_1
    assert(len(mass_map) == len(m_spring_k))
    if 'copies' in jdata :
        ncopies = np.prod(jdata['copies'])
    else :
        ncopies = 1

    with open(equi_conf) as f:
        sys_data = lmp.to_system_data(f.read().split('\n'))

    # sys_data = lmp.to_system_data(open(equi_conf).read().split('\n'))
    vol = np.linalg.det(sys_data['cell'])
    natoms = [ii * ncopies for ii in sys_data['atom_numbs']]
    
    Lambda_k = [compute_lambda(temp, ii) for ii in mass_map]
    Lambda_s = [compute_spring(temp, ii) for ii in m_spring_k]
    # print(np.log(Lambda_k), np.log(Lambda_s))
    
    with open(equi_conf) as fp:
        lines = list(fp)
    for idx,ii in enumerate(lines) :
        if 'Atoms' in ii :
            break
    first_type = int(lines[idx+2].split()[1]) - 1
    # print('# fixed atom of type %d ' % first_type)

    fe = 0
    fact = pc.Boltzmann * temp / pc.electron_volt / np.sum(natoms)
    for idx,ii in enumerate(natoms) :
        # kinetic contrib
        # print(idx)
        fe += 3 * ii * np.log(Lambda_k[idx])
        # print(3 * ii * np.log(Lambda_k[idx]) * fact)
        if (idx == first_type) :            
            fe += 3 * (ii-1) * np.log(Lambda_s[idx])
            fe += np.log(ii / (vol * (pc.angstrom**3)))
            # print(3.0 * (ii-1) * np.log(Lambda_s[idx]) * fact)
            # print(np.log(ii / (vol * (pc.angstrom**3))) * fact)
        else :
            fe += 3 * ii * np.log(Lambda_s[idx])
            # print(3.0 * ii * np.log(Lambda_s[idx]) * fact)
    fe *= pc.Boltzmann * temp / pc.electron_volt
    fe /= np.sum(natoms)
    return fe


def frenkel(job):
    with open(os.path.join(job, 'in.json')) as f:
        jdata = json.load(f)
    # jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))    
    equi_conf = jdata['equi_conf']
    cwd = os.getcwd()
    os.chdir(job)
    assert(os.path.isfile(equi_conf))
    equi_conf = os.path.abspath(equi_conf)
    os.chdir(cwd)
    temp = jdata['temp']
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    s_spring_k = jdata['spring_k']
    spring_k = jdata['spring_k']
    assert(type(spring_k) is not list)
    if type(spring_k) is not list:
        m_spring_k = []
        for ii in mass_map :
            m_spring_k.append(spring_k * ii)
        # spring_k = spring_k_1
    if 'copies' in jdata :
        ncopies = np.prod(jdata['copies'])
    else :
        ncopies = 1    
    with open(equi_conf) as f:
        sys_data = lmp.to_system_data(f.read().split('\n'))

    # sys_data = lmp.to_system_data(open(equi_conf).read().split('\n'))
    vol = np.linalg.det(sys_data['cell'])
    natoms = [ii * ncopies for ii in sys_data['atom_numbs']]

    Lambda_k = [compute_lambda(temp, ii) for ii in mass_map]
    Lambda_s = [compute_spring(temp, ii) for ii in m_spring_k]
    s_Lambda_s = compute_spring(temp, s_spring_k)

    fe = 0
    sum_m = 0
    fact = pc.Boltzmann * temp / pc.electron_volt / np.sum(natoms)
    for idx,ii in enumerate(natoms) :
        fe += 3.0 * ii * np.log(Lambda_k[idx])
        fe += 3.0 * ii * np.log(Lambda_s[idx])
        # print(idx)
        # print(3.0 * ii * np.log(Lambda_k[idx]) * fact)
        # print(3.0 * ii * np.log(Lambda_s[idx]) * fact)
        sum_m += mass_map[idx] * ii
    fe -= 3.0 * np.log(s_Lambda_s)
    fe -= 1.5 * np.log(sum_m)
    # fe += 2.0 * np.log(np.sum(natoms)/3.0)
    # print('# FS corr (does not apply)', 2.0 * np.log(np.sum(natoms)/3.0) *pc.Boltzmann * temp / pc.electron_volt / np.sum(natoms) * 3.0)
    # print((3.0 * np.log(s_Lambda_s) + 1.5 * np.log(sum_m)) * fact)
    fe += np.log(np.sum(natoms) / (vol * (pc.angstrom**3)))    
    # print(np.log(np.sum(natoms) / (vol * (pc.angstrom**3))) * fact, np.log(np.sum(natoms) / 3.0 / (vol * (pc.angstrom**3))) * fact)

    fe *= pc.Boltzmann * temp / pc.electron_volt
    fe /= np.sum(natoms)
    # total_atom_num = np.sum(natoms)
    # print('### debug:', fe, Lambda_k, Lambda_s, np.log(Lambda_k[0]), np.log(Lambda_s[0]))
    # print('### debug:average U', 3 * pc.Boltzmann * temp)
    # print('### debug:Helmholtz free energy', fe)
    # print('### debug:', (3*total_atom_num - 0.5 - 3*total_atom_num*np.log(Lambda_k[0]) - 3*(natoms[0]-1)*np.log(Lambda_s[0]) + np.log((vol * (pc.angstrom**3))) + 0.5*np.log(total_atom_num)) )
    return fe
        
    
def _main() :
    parser = argparse.ArgumentParser(
        description="Compute free energy of Einstein molecule")
    parser.add_argument('PARAM', type=str,
                        help='json parameter file')
    args = parser.parse_args()

    jdata = json.load(open(args.PARAM))
    fe = free_energy(jdata)
    print('# free energy of Einstein molecule in eV:')
    print(fe)

if __name__ == '__main__' :
    _main()



import os, json, shutil
import numpy as np
import unittest
from unittest.mock import MagicMock, patch, PropertyMock
from context import dpti
from dpti.lib.utils import get_file_md5, relative_link_file
from dpti.equi import post_task

class TestEquiMakeTask(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.mkdir('tmp_equi_log/')

    def setUp(self):
        self.maxDiff = None
        self.test_dir = 'tmp_equi_log'
        self.benchmark_dir = 'benchmark_equi_log'

    @patch('builtins.print')
    def test_npt(self, patch_print):
        test_name = 'npt'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)
        os.mkdir(test_dir)
        link_file_list = ['log.lammps', 'equi_settings.json', 'conf.lmp']
        for file in link_file_list:
            src = os.path.join(benchmark_dir, file)
            relative_link_file(file_path=src, target_dir=test_dir)

        post_task(test_dir)
        patch_print.assert_called_once()
        check_file_list = ['result', ]
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2))

    def test_get_npt_avg_conf(self):
        npt_name = os.path.join(self.benchmark_dir, 'npt')
        # test_dir = os.path.join(self.test_dir, 'get_npt_avg_conf')
        npt_avg_conf_lmp = dpti.equi.npt_equi_conf(npt_name)
        f2 = os.path.join(self.test_dir, 'npt_avg.lmp')
        with open(f2, 'w') as f:
            f.write(npt_avg_conf_lmp)
        f1 = os.path.join(npt_name, 'npt_avg.lmp')
        self.assertEqual(get_file_md5(f1), get_file_md5(f2))
    
    def test_compute_thermo(self):
        info = dpti.equi._compute_thermo('benchmark_equi_log/npt/log.lammps', 144, stat_skip=30, stat_bsize=10)
        with open('benchmark_equi_log/npt/result.json', 'r') as f:
            result = json.load(f)
        self.assertDictEqual(info, result)
        # print(info)
        # pass

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_equi_log/')


if __name__ == '__main__':
    unittest.main()
